from gymnasium.vector.vector_env import ArrayType, VectorEnv
import uuid
from typing import TYPE_CHECKING, Any, Generic, TypeVar

import numpy as np
import json
import gymnasium as gym
from gymnasium.core import ActType, ObsType, RenderFrame
from gymnasium.utils import seeding



from gymnasium.envs.registration import EnvSpec

ArrayType = TypeVar("ArrayType")

from orca_gym.environment.async_env.orca_gym_async_env import OrcaGymAsyncEnv

class OrcaGymVectorEnv(VectorEnv):
    """
    实现基于多个Agent的向量环境。
    这样可以用一个Mujoco Instance 来同时运行多个 env
    最后转换成VectorEnv的接口格式

    # ┌────────────┬────────────────┬──────────────────────┐
    # │            │ Mujoco Env     │ Robots in Mujoco     │
    # ├────────────┼────────────────┼──────────────────────┤
    # │ Vector Env │ OrcaGymAsyncEnv│ num_envs             │
    # │ RLLib      │ num_env_runners│ num_envs_per_runner  │
    # │ LeggedGym  │ subenv_num     │ agent_num            │
    # └────────────┴────────────────┴──────────────────────┘
    
    
    Base class for vectorized environments to run multiple independent copies of the same environment in parallel.

    Vector environments can provide a linear speed-up in the steps taken per second through sampling multiple
    sub-environments at the same time. Gymnasium contains two generalised Vector environments: :class:`AsyncVectorEnv`
    and :class:`SyncVectorEnv` along with several custom vector environment implementations.
    For :func:`reset` and :func:`step` batches `observations`, `rewards`,  `terminations`, `truncations` and
    `info` for each sub-environment, see the example below. For the `rewards`, `terminations`, and `truncations`,
    the data is packaged into a NumPy array of shape `(num_envs,)`. For `observations` (and `actions`, the batching
    process is dependent on the type of observation (and action) space, and generally optimised for neural network
    input/outputs. For `info`, the data is kept as a dictionary such that a key will give the data for all sub-environment.

    For creating environments, :func:`make_vec` is a vector environment equivalent to :func:`make` for easily creating
    vector environments that contains several unique arguments for modifying environment qualities, number of environment,
    vectorizer type, vectorizer arguments.

    Note:
        The info parameter of :meth:`reset` and :meth:`step` was originally implemented before v0.25 as a list
        of dictionary for each sub-environment. However, this was modified in v0.25+ to be a dictionary with a NumPy
        array for each key. To use the old info style, utilise the :class:`DictInfoToList` wrapper.

    To avoid having to wait for all sub-environments to terminated before resetting, implementations will autoreset
    sub-environments on episode end (`terminated or truncated is True`). As a result, when adding observations
    to a replay buffer, this requires knowing when an observation (and info) for each sub-environment are the first
    observation from an autoreset. We recommend using an additional variable to store this information such as
    ``has_autoreset = np.logical_or(terminated, truncated)``.

    The Vector Environments have the additional attributes for users to understand the implementation

    - :attr:`num_envs` - The number of sub-environment in the vector environment
    - :attr:`observation_space` - The batched observation space of the vector environment
    - :attr:`single_observation_space` - The observation space of a single sub-environment
    - :attr:`action_space` - The batched action space of the vector environment
    - :attr:`single_action_space` - The action space of a single sub-environment
    """

    metadata: dict[str, Any] = {}
    spec: EnvSpec | None = None
    render_mode: str | None = None
    closed: bool = False

    observation_space: gym.Space
    action_space: gym.Space
    single_observation_space: gym.Space
    single_action_space: gym.Space

    num_envs: int

    _np_random: np.random.Generator | None = None
    _np_random_seed: int | None = None

    def __init__(self, 
        num_envs: int, 
        worker_index: int,
        entry_point: str,
        **kwargs
    ):
        self.agent_num = num_envs
        assert num_envs % 32 == 0, "num_envs must be a multiple of 32"
        self.env_num = num_envs // 32

        self.worker_index = worker_index
        env_id = kwargs.get("env_id", "")
        env_id_prefix = "-".join(env_id.split("-")[:-1])

        self.envs = []
        for i in range(self.env_num):
            worker_env_id = f"{env_id_prefix}-{i:03d}-{worker_index:03d}"
            kwargs["env_id"] = worker_env_id
            if worker_index == 1 and i == 0:
                kwargs["is_subenv"] = False
            else:
                kwargs["is_subenv"] = True

            # print("Create OrcaGymVectorEnv: env_id={}, worker_index={}, kwargs={}, entry_point={}".format(worker_env_id, worker_index, kwargs, entry_point))

            gym.register(
                id=worker_env_id,
                entry_point=entry_point,
                kwargs=kwargs
            )

            self.envs.append(gym.make(worker_env_id, **kwargs))

        # OrcaGymAsyncEnv 的 obs 是 字典形式，不符合rllib的格式，需要转换为np形式
        obs_list = []
        for env in self.envs:
            unwrapped_env : OrcaGymAsyncEnv = env.unwrapped
            env_obs, _, _, _ = unwrapped_env.get_obs()
            obs_list.append(env_obs["observation"])

        # 将所有 env 的 obs 合并成一个 np.ndarray
        obs = np.concatenate(obs_list, axis=0)
        self.observation_space = unwrapped_env.generate_observation_space(obs)
        self.single_observation_space = unwrapped_env.generate_observation_space(obs[0])

        self.single_action_space = self.envs[0].action_space
        # 将 single_action_space 扩展到 env_num 维度
        self.action_space = gym.spaces.Box(
            low=np.tile(self.single_action_space.low, (self.agent_num, 1)),
            high=np.tile(self.single_action_space.high, (self.agent_num, 1)),
            dtype=self.single_action_space.dtype,
        )

        # print("OrcaGymVectorEnv observation_space: ", self.observation_space)
        # print("OrcaGymVectorEnv single_observation_space: ", self.single_observation_space)
        # print("OrcaGymVectorEnv action_space: ", self.action_space)
        # print("OrcaGymVectorEnv single_action_space: ", self.single_action_space)

        self.num_envs = num_envs
        self.closed = False
        self._np_random = None

    def reset(
        self,
        *,
        seed: int | None = None,
        options: dict[str, Any] | None = None,
    ) -> tuple[ObsType, dict[str, Any]]:  # type: ignore
        """Reset all parallel environments and return a batch of initial observations and info.

        Args:
            seed: The environment reset seed
            options: If to return the options

        Returns:
            A batch of observations and info from the vectorized environment.

        Example:
            >>> import gymnasium as gym
            >>> envs = gym.make_vec("CartPole-v1", num_envs=3, vectorization_mode="sync")
            >>> observations, infos = envs.reset(seed=42)
            >>> observations
            array([[ 0.0273956 , -0.00611216,  0.03585979,  0.0197368 ],
                   [ 0.01522993, -0.04562247, -0.04799704,  0.03392126],
                   [-0.03774345, -0.02418869, -0.00942293,  0.0469184 ]],
                  dtype=float32)
            >>> infos
            {}
        """
        if seed is not None:
            self._np_random, self._np_random_seed = seeding.np_random(seed)

        # obs 只取第一个 agent 的观测数据，实际所有 agent 的观测数据在 info 中
        obs_list = []
        for env in self.envs:
            _, async_env_info = env.reset()
            obs_list.append(async_env_info["env_obs"]["observation"])

        obs = np.concatenate(obs_list, axis=0)
        infos = [{} for _ in range(self.agent_num)]

        # print("OrcaGymVectorEnv reset obs shape: ", obs.shape)

        return obs, infos

    def step(
        self, actions: ActType
    ) -> tuple[ObsType, ArrayType, ArrayType, ArrayType, dict[str, Any]]:
        """Take an action for each parallel environment.

        Args:
            actions: Batch of actions with the :attr:`action_space` shape.

        Returns:
            Batch of (observations, rewards, terminations, truncations, infos)

        Note:
            As the vector environments autoreset for a terminating and truncating sub-environments, this will occur on
            the next step after `terminated or truncated is True`.

        Example:
            >>> import gymnasium as gym
            >>> import numpy as np
            >>> envs = gym.make_vec("CartPole-v1", num_envs=3, vectorization_mode="sync")
            >>> _ = envs.reset(seed=42)
            >>> actions = np.array([1, 0, 1], dtype=np.int32)
            >>> observations, rewards, terminations, truncations, infos = envs.step(actions)
            >>> observations
            array([[ 0.02727336,  0.18847767,  0.03625453, -0.26141977],
                   [ 0.01431748, -0.24002443, -0.04731862,  0.3110827 ],
                   [-0.03822722,  0.1710671 , -0.00848456, -0.2487226 ]],
                  dtype=float32)
            >>> rewards
            array([1., 1., 1.])
            >>> terminations
            array([False, False, False])
            >>> terminations
            array([False, False, False])
            >>> infos
            {}
        """
        # raise NotImplementedError(f"{self.__str__()} step function is not implemented.")
        # print("OrcaGymVectorEnv step actions shape: ", actions.shape, "worker_index: ", self.worker_index, "actions: ", actions)

        # 假设原数组 action_array 形状为 (N, 12)，N 是第一维大小
        N = actions.shape[0]  # 获取第一维长度

        # 计算整除64后的分组数量
        num_groups = N // 32  # 整数除法（自动舍去余数）

        # 重塑为 (num_groups, 32, 12) 形状
        reshaped_action = actions.reshape(num_groups, 32, actions.shape[1])

        obs_list = []
        reward_list = []
        terminated_list = []
        truncated_list = []
        for i, env in enumerate(self.envs):
            _, _, _, _, info = env.step(reshaped_action[i].flatten())
            obs_list.append(info["env_obs"]["observation"])
            reward_list.append(info["reward"])
            terminated_list.append(info["terminated"])
            truncated_list.append(info["truncated"])

        obs = np.concatenate(obs_list, axis=0)
        reward = np.concatenate(reward_list, axis=0)
        terminated = np.concatenate(terminated_list, axis=0)
        truncated = np.concatenate(truncated_list, axis=0)
        infos = [{} for _ in range(self.agent_num)]

        # print("OrcaGymVectorEnv step obs shape: ", obs.shape)
        # print("OrcaGymVectorEnv step reward shape: ", reward.shape)
        # print("OrcaGymVectorEnv step terminated: ", terminated)
        # print("OrcaGymVectorEnv step truncated: ", truncated)

        return obs, reward, terminated, truncated, infos

    def render(self) -> tuple[RenderFrame, ...] | None:
        """Returns the rendered frames from the parallel environments.

        Returns:
            A tuple of rendered frames from the parallel environments
        """

        # Render 不需要都调用，反正只有第一个需要执行Render
        self.envs[0].render()

    def close(self, **kwargs: Any):
        """Close all parallel environments and release resources.

        It also closes all the existing image viewers, then calls :meth:`close_extras` and set
        :attr:`closed` as ``True``.

        Warnings:
            This function itself does not close the environments, it should be handled
            in :meth:`close_extras`. This is generic for both synchronous and asynchronous
            vectorized environments.

        Note:
            This will be automatically called when garbage collected or program exited.

        Args:
            **kwargs: Keyword arguments passed to :meth:`close_extras`
        """
        if self.closed:
            return

        for env in self.envs:
            env.close()

        self.close_extras(**kwargs)
        self.closed = True

    def _unpack_reset_info(self, info: dict[str, Any]) -> dict[str, Any]:
        return info["env_obs"]

    def _unpack_step_info(self, info: dict[str, Any]) -> dict[str, Any]:
        return info["env_obs"]


    # def close_extras(self, **kwargs: Any):
    #     """Clean up the extra resources e.g. beyond what's in this base class."""
    #     pass

    # @property
    # def np_random(self) -> np.random.Generator:
    #     """Returns the environment's internal :attr:`_np_random` that if not set will initialise with a random seed.

    #     Returns:
    #         Instances of `np.random.Generator`
    #     """
    #     if self._np_random is None:
    #         self._np_random, self._np_random_seed = seeding.np_random()
    #     return self._np_random

    # @np_random.setter
    # def np_random(self, value: np.random.Generator):
    #     self._np_random = value
    #     self._np_random_seed = -1

    # @property
    # def np_random_seed(self) -> int | None:
    #     """Returns the environment's internal :attr:`_np_random_seed` that if not set will first initialise with a random int as seed.

    #     If :attr:`np_random_seed` was set directly instead of through :meth:`reset` or :meth:`set_np_random_through_seed`,
    #     the seed will take the value -1.

    #     Returns:
    #         int: the seed of the current `np_random` or -1, if the seed of the rng is unknown
    #     """
    #     if self._np_random_seed is None:
    #         self._np_random, self._np_random_seed = seeding.np_random()
    #     return self._np_random_seed

    # @property
    # def unwrapped(self):
    #     """Return the base environment."""
    #     return self

    # def _add_info(
    #     self, vector_infos: dict[str, Any], env_info: dict[str, Any], env_num: int
    # ) -> dict[str, Any]:
    #     """Add env info to the info dictionary of the vectorized environment.

    #     Given the `info` of a single environment add it to the `infos` dictionary
    #     which represents all the infos of the vectorized environment.
    #     Every `key` of `info` is paired with a boolean mask `_key` representing
    #     whether or not the i-indexed environment has this `info`.

    #     Args:
    #         vector_infos (dict): the infos of the vectorized environment
    #         env_info (dict): the info coming from the single environment
    #         env_num (int): the index of the single environment

    #     Returns:
    #         infos (dict): the (updated) infos of the vectorized environment
    #     """
    #     for key, value in env_info.items():
    #         # If value is a dictionary, then we apply the `_add_info` recursively.
    #         if isinstance(value, dict):
    #             array = self._add_info(vector_infos.get(key, {}), value, env_num)
    #         # Otherwise, we are a base case to group the data
    #         else:
    #             # If the key doesn't exist in the vector infos, then we can create an array of that batch type
    #             if key not in vector_infos:
    #                 if type(value) in [int, float, bool] or issubclass(
    #                     type(value), np.number
    #                 ):
    #                     array = np.zeros(self.num_envs, dtype=type(value))
    #                 elif isinstance(value, np.ndarray):
    #                     # We assume that all instances of the np.array info are of the same shape
    #                     array = np.zeros(
    #                         (self.num_envs, *value.shape), dtype=value.dtype
    #                     )
    #                 else:
    #                     # For unknown objects, we use a Numpy object array
    #                     array = np.full(self.num_envs, fill_value=None, dtype=object)
    #             # Otherwise, just use the array that already exists
    #             else:
    #                 array = vector_infos[key]

    #             # Assign the data in the `env_num` position
    #             #   We only want to run this for the base-case data (not recursive data forcing the ugly function structure)
    #             array[env_num] = value

    #         # Get the array mask and if it doesn't already exist then create a zero bool array
    #         array_mask = vector_infos.get(
    #             f"_{key}", np.zeros(self.num_envs, dtype=np.bool_)
    #         )
    #         array_mask[env_num] = True

    #         # Update the vector info with the updated data and mask information
    #         vector_infos[key], vector_infos[f"_{key}"] = array, array_mask

    #     return vector_infos

    # def __del__(self):
    #     """Closes the vector environment."""
    #     if not getattr(self, "closed", True):
    #         self.close()

    # def __repr__(self) -> str:
    #     """Returns a string representation of the vector environment.

    #     Returns:
    #         A string containing the class name, number of environments and environment spec id
    #     """
    #     if self.spec is None:
    #         return f"{self.__class__.__name__}(num_envs={self.num_envs})"
    #     else:
    #         return (
    #             f"{self.__class__.__name__}({self.spec.id}, num_envs={self.num_envs})"
    #         )


from .orca_gym_env import OrcaGymBaseEnv, RewardType
from .orca_gym_local_env import OrcaGymLocalEnv
from .orca_gym_remote_env import OrcaGymRemoteEnv


__all__ = ['OrcaGymBaseEnv', 'RewardType', 'OrcaGymLocalEnv', 'OrcaGymRemoteEnv']


from collections import defaultdict
from functools import partial
import logging
import math
import time
from typing import Collection, DefaultDict, List, Optional, Union

import gymnasium as gym
import ray
from gymnasium.wrappers.vector import DictInfoToList

from ray.rllib.algorithms.algorithm_config import AlgorithmConfig
from ray.rllib.callbacks.callbacks import RLlibCallback
from ray.rllib.callbacks.utils import make_callback
from ray.rllib.core import (
    COMPONENT_ENV_TO_MODULE_CONNECTOR,
    COMPONENT_MODULE_TO_ENV_CONNECTOR,
    COMPONENT_RL_MODULE,
    DEFAULT_AGENT_ID,
    DEFAULT_MODULE_ID,
)
from ray.rllib.core.columns import Columns
from ray.rllib.core.rl_module.rl_module import RLModule, RLModuleSpec
from ray.rllib.env import INPUT_ENV_SPACES, INPUT_ENV_SINGLE_SPACES
from ray.rllib.env.env_context import EnvContext
from ray.rllib.env.env_runner import EnvRunner, ENV_STEP_FAILURE
from ray.rllib.env.single_agent_episode import SingleAgentEpisode
from ray.rllib.env.utils import _gym_env_creator
from ray.rllib.utils import force_list
from ray.rllib.utils.annotations import override
from ray.rllib.utils.checkpoints import Checkpointable
from ray.rllib.utils.deprecation import Deprecated
from ray.rllib.utils.framework import get_device
from ray.rllib.utils.metrics import (
    ENV_TO_MODULE_CONNECTOR,
    EPISODE_DURATION_SEC_MEAN,
    EPISODE_LEN_MAX,
    EPISODE_LEN_MEAN,
    EPISODE_LEN_MIN,
    EPISODE_RETURN_MAX,
    EPISODE_RETURN_MEAN,
    EPISODE_RETURN_MIN,
    MODULE_TO_ENV_CONNECTOR,
    NUM_AGENT_STEPS_SAMPLED,
    NUM_AGENT_STEPS_SAMPLED_LIFETIME,
    NUM_ENV_STEPS_SAMPLED,
    NUM_ENV_STEPS_SAMPLED_LIFETIME,
    NUM_EPISODES,
    NUM_EPISODES_LIFETIME,
    NUM_MODULE_STEPS_SAMPLED,
    NUM_MODULE_STEPS_SAMPLED_LIFETIME,
    RLMODULE_INFERENCE_TIMER,
    SAMPLE_TIMER,
    TIME_BETWEEN_SAMPLING,
    WEIGHTS_SEQ_NO,
)
from ray.rllib.utils.spaces.space_utils import unbatch
from ray.rllib.utils.typing import EpisodeID, ResultDict, StateDict
from ray.tune.registry import ENV_CREATOR, _global_registry
from ray.util.annotations import PublicAPI

logger = logging.getLogger("ray.rllib")

from ray.rllib.env.single_agent_env_runner import SingleAgentEnvRunner
from orca_gym.environment.async_env.orca_gym_vector_env import OrcaGymVectorEnv

class OrcaGymAsyncSingleAgentEnvRunner(SingleAgentEnvRunner):
    """
    Wrap the SingleAgentEnvRunner to support orca gym asynchronous environments.
    """

    @override(SingleAgentEnvRunner)
    def __init__(self, *, config: AlgorithmConfig, **kwargs):
        """Initializes a OrcaGymAsyncSingleAgentEnvRunner instance.

        Args:
            config: An `AlgorithmConfig` object containing all settings needed to
                build this `EnvRunner` class.
        """
        super().__init__(config=config, **kwargs)

        print("env_runner.module is : ", self.module)

       

    @override(SingleAgentEnvRunner)
    def make_env(self) -> None:
        """Creates a vectorized gymnasium env and stores it in `self.env`.

        Note that users can change the EnvRunner's config (e.g. change
        `self.config.env_config`) and then call this method to create new environments
        with the updated configuration.
        """
        # If an env already exists, try closing it first (to allow it to properly
        # cleanup).
        if self.env is not None:
            try:
                self.env.close()
            except Exception as e:
                logger.warning(
                    "Tried closing the existing env, but failed with error: "
                    f"{e.args[0]}"
                )

        env_ctx = self.config.env_config
        if not isinstance(env_ctx, EnvContext):
            env_ctx = EnvContext(
                env_ctx,
                worker_index=self.worker_index,
                num_workers=self.num_workers,
                remote=self.config.remote_worker_envs,
            )

        # No env provided -> Error.
        if not self.config.env:
            raise ValueError(
                "`config.env` is not provided! You should provide a valid environment "
                "to your config through `config.environment([env descriptor e.g. "
                "'CartPole-v1'])`."
            )
        # Register env for the local context.
        # Note, `gym.register` has to be called on each worker.
        elif isinstance(self.config.env, str) and _global_registry.contains(
            ENV_CREATOR, self.config.env
        ):
            entry_point = partial(
                _global_registry.get(ENV_CREATOR, self.config.env),
                env_ctx,
            )
        else:
            entry_point = partial(
                _gym_env_creator,
                env_descriptor=self.config.env,
                env_context=env_ctx,
            )
        gym.register("rllib-single-agent-env-v0", entry_point=entry_point)
        vectorize_mode = self.config.gym_env_vectorize_mode

        # print(f"Make vec env: config.env type={type(self.config.env)}, env= {self.config.env}, num_envs_per_env_runner= {self.config.num_envs_per_env_runner}, worker_index= {self.worker_index}")
        env_kwargs = self.config.env_config.get("env_kwargs", {})
        entry_point = self.config.env_config.get("entry_point", "")
        # print("Make vec env: env_kwargs: ", env_kwargs)
        self.env = OrcaGymVectorEnv(
            num_envs=self.config.num_envs_per_env_runner,
            worker_index=self.worker_index,
            entry_point=entry_point,
            **env_kwargs
        )


        # self.env = DictInfoToList(
        #     gym.make_vec(
        #         "rllib-single-agent-env-v0",
        #         num_envs=self.config.num_envs_per_env_runner,
        #         vectorization_mode=(
        #             vectorize_mode
        #             if isinstance(vectorize_mode, gym.envs.registration.VectorizeMode)
        #             else gym.envs.registration.VectorizeMode(vectorize_mode.lower())
        #         ),
        #     )
        # )

        self.num_envs: int = self.env.num_envs
        assert self.num_envs == self.config.num_envs_per_env_runner

        # Set the flag to reset all envs upon the next `sample()` call.
        self._needs_initial_reset = True

        # Call the `on_environment_created` callback.
        make_callback(
            "on_environment_created",
            callbacks_objects=self._callbacks,
            callbacks_functions=self.config.callbacks_on_environment_created,
            kwargs=dict(
                env_runner=self,
                metrics_logger=self.metrics,
                env=self.env.unwrapped,
                env_context=env_ctx,
            ),
        )




import numpy as np
from orca_gym.scene.orca_gym_scene import OrcaGymScene
from orca_gym.scene.orca_gym_scene import Actor, LightInfo, CameraSensorInfo, MaterialInfo


class OrcaGymSceneRuntime():
    """
    将 OrcaGymScene 的功能封装为一个运行时类
    Env 对象可以持有Runtime类的实例，实现部分控制Scene的能力。

    - 允许的Scene操作：
        例如 调整 Light 属性， 切换 Camera。这类操作不涉及场景资源销毁和仿真对象的变更

    - 禁止的Scene操作：
        例如 添加 Actor， Publish Scene。这类操作涉及资源销毁和仿真对象的变更，无法在仿真运行过程中使用
    """

    def __init__(
        self,
        scene: OrcaGymScene
    ):
        # Initialize the scene
        self.scene = scene

    def set_light_info(self, actor_name : str, light_info: LightInfo):
        try:
            self.scene.set_light_info(actor_name, light_info)
        except Exception as e:
            print(f"Failed to set light info for {actor_name}: {e}")


    def make_camera_viewport_active(self, actor_name: str, entity_name: str):
        try:
            self.scene.make_camera_viewport_active(actor_name, entity_name)
        except Exception as e:
            print(f"Failed to make camera viewport active for {actor_name}: {e}")

    def set_actor_anim_param_number(self, actor_name: str, param_name: str, value: float):
        try:
            self.scene.set_actor_anim_param_number(actor_name, param_name, value)
        except Exception as e:
            print(f"Failed to set actor anim param number for {actor_name}: {e}")

    def set_actor_anim_param_bool(self, actor_name: str, param_name: str, value: bool):
        try:
            self.scene.set_actor_anim_param_bool(actor_name, param_name, value)
        except Exception as e:
            print(f"Failed to set actor anim param bool for {actor_name}: {e}")

    def set_actor_anim_param_string(self, actor_name: str, param_name: str, value: str):
        try:
            self.scene.set_actor_anim_param_string(actor_name, param_name, value)
        except Exception as e:
            print(f"Failed to set actor anim param string for {actor_name}: {e}")


from typing import Dict, Tuple, Union

import numpy as np
from gymnasium import error

try:
    import mujoco
    from mujoco import MjData, MjModel, mjtObj
except ImportError as e:
    raise error.DependencyNotInstalled(f"{e}. (HINT: you need to install mujoco")

MJ_OBJ_TYPES = [
    "mjOBJ_BODY",
    "mjOBJ_JOINT",
    "mjOBJ_GEOM",
    "mjOBJ_SITE",
    "mjOBJ_CAMERA",
    "mjOBJ_ACTUATOR",
    "mjOBJ_SENSOR",
]


def robot_get_obs(model, data, joint_names):
    """Returns all joint positions and velocities associated with a robot."""
    if data.qpos is not None and joint_names:
        names = [n for n in joint_names if n.startswith("robot")]
        return (
            np.squeeze(np.array([get_joint_qpos(model, data, name) for name in names])),
            np.squeeze(np.array([get_joint_qvel(model, data, name) for name in names])),
        )
    return np.zeros(0), np.zeros(0)


def ctrl_set_action(model, data, action):
    """For torque actuators it copies the action into mujoco ctrl field.

    For position actuators it sets the target relative to the current qpos.
    """
    if model.nmocap > 0:
        _, action = np.split(action, (model.nmocap * 7,))

    if len(data.ctrl) > 0:
        for i in range(action.shape[0]):
            if model.actuator_biastype[i] == 0:
                data.ctrl[i] = action[i]
            else:
                idx = model.jnt_qposadr[model.actuator_trnid[i, 0]]
                data.ctrl[i] = data.qpos[idx] + action[i]


def mocap_set_action(model, data, action):
    """Update the position of the mocap body with the desired action.

    The action controls the robot using mocaps. Specifically, bodies
    on the robot (for example the gripper wrist) is controlled with
    mocap bodies. In this case the action is the desired difference
    in position and orientation (quaternion), in world coordinates,
    of the target body. The mocap is positioned relative to
    the target body according to the delta, and the MuJoCo equality
    constraint optimizer tries to center the welded body on the mocap.
    """
    if model.nmocap > 0:
        action, _ = np.split(action, (model.nmocap * 7,))
        action = action.reshape(model.nmocap, 7)

        pos_delta = action[:, :3]
        quat_delta = action[:, 3:]

        reset_mocap2body_xpos(model, data)
        data.mocap_pos[:] = data.mocap_pos + pos_delta
        data.mocap_quat[:] = data.mocap_quat + quat_delta


def reset_mocap_welds(model, data):
    """Resets the mocap welds that we use for actuation."""
    if model.nmocap > 0 and model.eq_data is not None:
        for i in range(model.eq_data.shape[0]):
            if model.eq_type[i] == mujoco.mjtEq.mjEQ_WELD:
                model.eq_data[i, :7] = np.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0])
    mujoco.mj_forward(model, data)


def reset_mocap2body_xpos(model, data):
    """Resets the position and orientation of the mocap bodies to the same
    values as the bodies they're welded to.
    """

    if model.eq_type is None or model.eq_obj1id is None or model.eq_obj2id is None:
        return
    for eq_type, obj1_id, obj2_id in zip(
        model.eq_type, model.eq_obj1id, model.eq_obj2id
    ):
        if eq_type != mujoco.mjtEq.mjEQ_WELD:
            continue

        mocap_id = model.body_mocapid[obj1_id]
        if mocap_id != -1:
            # obj1 is the mocap, obj2 is the welded body
            body_idx = obj2_id
        else:
            # obj2 is the mocap, obj1 is the welded body
            mocap_id = model.body_mocapid[obj2_id]
            body_idx = obj1_id

        assert mocap_id != -1
        data.mocap_pos[mocap_id][:] = data.xpos[body_idx]
        data.mocap_quat[mocap_id][:] = data.xquat[body_idx]


def get_site_jacp(model, data, site_id):
    """Return the Jacobian' translational component of the end-effector of
    the corresponding site id.
    """
    jacp = np.zeros((3, model.nv))
    mujoco.mj_jacSite(model, data, jacp, None, site_id)

    return jacp


def get_site_jacr(model, data, site_id):
    """Return the Jacobian' rotational component of the end-effector of
    the corresponding site id.
    """
    jacr = np.zeros((3, model.nv))
    mujoco.mj_jacSite(model, data, None, jacr, site_id)

    return jacr


def set_joint_qpos(model, data, name, value):
    """Set the joint positions (qpos) of the model."""
    joint_id = mujoco.mj_name2id(model, mujoco.mjtObj.mjOBJ_JOINT, name)
    joint_type = model.jnt_type[joint_id]
    joint_addr = model.jnt_qposadr[joint_id]

    if joint_type == mujoco.mjtJoint.mjJNT_FREE:
        ndim = 7
    elif joint_type == mujoco.mjtJoint.mjJNT_BALL:
        ndim = 4
    else:
        assert joint_type in (mujoco.mjtJoint.mjJNT_HINGE, mujoco.mjtJoint.mjJNT_SLIDE)
        ndim = 1

    start_idx = joint_addr
    end_idx = joint_addr + ndim
    value = np.array(value)
    if ndim > 1:
        assert value.shape == (
            end_idx - start_idx
        ), f"Value has incorrect shape {name}: {value}"
    data.qpos[start_idx:end_idx] = value


def set_joint_qvel(model, data, name, value):
    """Set the joints linear and angular (qvel) of the model."""
    joint_id = mujoco.mj_name2id(model, mujoco.mjtObj.mjOBJ_JOINT, name)
    joint_type = model.jnt_type[joint_id]
    joint_addr = model.jnt_dofadr[joint_id]

    if joint_type == mujoco.mjtJoint.mjJNT_FREE:
        ndim = 6
    elif joint_type == mujoco.mjtJoint.mjJNT_BALL:
        ndim = 3
    else:
        assert joint_type in (mujoco.mjtJoint.mjJNT_HINGE, mujoco.mjtJoint.mjJNT_SLIDE)
        ndim = 1

    start_idx = joint_addr
    end_idx = joint_addr + ndim
    value = np.array(value)
    if ndim > 1:
        assert value.shape == (
            end_idx - start_idx
        ), f"Value has incorrect shape {name}: {value}"
    data.qvel[start_idx:end_idx] = value


def get_joint_qpos(model, data, name):
    """Return the joints position and orientation (qpos) of the model."""
    joint_id = mujoco.mj_name2id(model, mujoco.mjtObj.mjOBJ_JOINT, name)
    joint_type = model.jnt_type[joint_id]
    joint_addr = model.jnt_qposadr[joint_id]

    if joint_type == mujoco.mjtJoint.mjJNT_FREE:
        ndim = 7
    elif joint_type == mujoco.mjtJoint.mjJNT_BALL:
        ndim = 4
    else:
        assert joint_type in (mujoco.mjtJoint.mjJNT_HINGE, mujoco.mjtJoint.mjJNT_SLIDE)
        ndim = 1

    start_idx = joint_addr
    end_idx = joint_addr + ndim

    return data.qpos[start_idx:end_idx].copy()


def get_joint_qvel(model, data, name):
    """Return the joints linear and angular velocities (qvel) of the model."""
    joint_id = mujoco.mj_name2id(model, mujoco.mjtObj.mjOBJ_JOINT, name)
    joint_type = model.jnt_type[joint_id]
    joint_addr = model.jnt_dofadr[joint_id]

    if joint_type == mujoco.mjtJoint.mjJNT_FREE:
        ndim = 6
    elif joint_type == mujoco.mjtJoint.mjJNT_BALL:
        ndim = 4
    else:
        assert joint_type in (mujoco.mjtJoint.mjJNT_HINGE, mujoco.mjtJoint.mjJNT_SLIDE)
        ndim = 1

    start_idx = joint_addr
    end_idx = joint_addr + ndim

    return data.qvel[start_idx:end_idx].copy()


def get_site_xpos(model, data, name):
    site_id = mujoco.mj_name2id(model, mujoco.mjtObj.mjOBJ_SITE, name)
    return data.site_xpos[site_id]


def get_site_xvelp(model, data, name):
    site_id = mujoco.mj_name2id(model, mujoco.mjtObj.mjOBJ_SITE, name)
    jacp = get_site_jacp(model, data, site_id)
    xvelp = jacp @ data.qvel
    return xvelp


def get_site_xvelr(model, data, name):
    site_id = mujoco.mj_name2id(model, mujoco.mjtObj.mjOBJ_SITE, name)
    jacp = get_site_jacr(model, data, site_id)
    xvelp = jacp @ data.qvel
    return xvelp


def set_mocap_pos(model, data, name, value):
    body_id = mujoco.mj_name2id(model, mujoco.mjtObj.mjOBJ_BODY, name)
    mocap_id = model.body_mocapid[body_id]
    data.mocap_pos[mocap_id] = value


def set_mocap_quat(model: MjModel, data: MjData, name: str, value):
    body_id = mujoco.mj_name2id(model, mujoco.mjtObj.mjOBJ_BODY, name)
    mocap_id = model.body_mocapid[body_id]
    data.mocap_quat[mocap_id] = value


def get_site_xmat(model: MjModel, data: MjData, name: str):
    site_id = mujoco.mj_name2id(model, mujoco.mjtObj.mjOBJ_SITE, name)
    return data.site_xmat[site_id].reshape(3, 3)


def extract_mj_names(
    model: MjModel, obj_type: mjtObj
) -> Tuple[Union[Tuple[str, ...], Tuple[()]], Dict[str, int], Dict[int, str]]:

    if obj_type == mujoco.mjtObj.mjOBJ_BODY:
        name_addr = model.name_bodyadr
        n_obj = model.nbody

    elif obj_type == mujoco.mjtObj.mjOBJ_JOINT:
        name_addr = model.name_jntadr
        n_obj = model.njnt

    elif obj_type == mujoco.mjtObj.mjOBJ_GEOM:
        name_addr = model.name_geomadr
        n_obj = model.ngeom

    elif obj_type == mujoco.mjtObj.mjOBJ_SITE:
        name_addr = model.name_siteadr
        n_obj = model.nsite

    elif obj_type == mujoco.mjtObj.mjOBJ_LIGHT:
        name_addr = model.name_lightadr
        n_obj = model.nlight

    elif obj_type == mujoco.mjtObj.mjOBJ_CAMERA:
        name_addr = model.name_camadr
        n_obj = model.ncam

    elif obj_type == mujoco.mjtObj.mjOBJ_ACTUATOR:
        name_addr = model.name_actuatoradr
        n_obj = model.nu

    elif obj_type == mujoco.mjtObj.mjOBJ_SENSOR:
        name_addr = model.name_sensoradr
        n_obj = model.nsensor

    elif obj_type == mujoco.mjtObj.mjOBJ_TENDON:
        name_addr = model.name_tendonadr
        n_obj = model.ntendon

    elif obj_type == mujoco.mjtObj.mjOBJ_MESH:
        name_addr = model.name_meshadr
        n_obj = model.nmesh
    else:
        raise ValueError(
            "`{}` was passed as the MuJoCo model object type. The MuJoCo model object type can only be of the following mjtObj enum types: {}.".format(
                obj_type, MJ_OBJ_TYPES
            )
        )

    id2name = {i: None for i in range(n_obj)}
    name2id = {}
    for addr in name_addr:
        name = model.names[addr:].split(b"\x00")[0].decode()
        if name:
            obj_id = mujoco.mj_name2id(model, obj_type, name)
            assert 0 <= obj_id < n_obj and id2name[obj_id] is None
            name2id[name] = obj_id
            id2name[obj_id] = name

    return tuple(id2name[id] for id in sorted(name2id.values())), name2id, id2name


class MujocoModelNames:
    """Access mjtObj object names and ids of the current MuJoCo model.

    This class supports access to the names and ids of the following mjObj types:
        mjOBJ_BODY
        mjOBJ_JOINT
        mjOBJ_GEOM
        mjOBJ_SITE
        mjOBJ_CAMERA
        mjOBJ_ACTUATOR
        mjOBJ_SENSOR

    The properties provided for each ``mjObj`` are:
        ``mjObj``_names: list of the mjObj names in the model of type mjOBJ_FOO.
        ``mjObj``_name2id: dictionary with name of the mjObj as keys and id of the mjObj as values.
        ``mjObj``_id2name: dictionary with id of the mjObj as keys and name of the mjObj as values.
    """

    def __init__(self, model: MjModel):
        """Access mjtObj object names and ids of the current MuJoCo model.

        Args:
            model: mjModel of the MuJoCo environment.
        """
        (
            self._body_names,
            self._body_name2id,
            self._body_id2name,
        ) = extract_mj_names(model, mujoco.mjtObj.mjOBJ_BODY)
        (
            self._joint_names,
            self._joint_name2id,
            self._joint_id2name,
        ) = extract_mj_names(model, mujoco.mjtObj.mjOBJ_JOINT)
        (
            self._geom_names,
            self._geom_name2id,
            self._geom_id2name,
        ) = extract_mj_names(model, mujoco.mjtObj.mjOBJ_GEOM)
        (
            self._site_names,
            self._site_name2id,
            self._site_id2name,
        ) = extract_mj_names(model, mujoco.mjtObj.mjOBJ_SITE)
        (
            self._camera_names,
            self._camera_name2id,
            self._camera_id2name,
        ) = extract_mj_names(model, mujoco.mjtObj.mjOBJ_CAMERA)
        (
            self._actuator_names,
            self._actuator_name2id,
            self._actuator_id2name,
        ) = extract_mj_names(model, mujoco.mjtObj.mjOBJ_ACTUATOR)
        (
            self._sensor_names,
            self._sensor_name2id,
            self._sensor_id2name,
        ) = extract_mj_names(model, mujoco.mjtObj.mjOBJ_SENSOR)

    @property
    def body_names(self):
        return self._body_names

    @property
    def body_name2id(self):
        return self._body_name2id

    @property
    def body_id2name(self):
        return self._body_id2name

    @property
    def joint_names(self):
        return self._joint_names

    @property
    def joint_name2id(self):
        return self._joint_name2id

    @property
    def joint_id2name(self):
        return self._joint_id2name

    @property
    def geom_names(self):
        return self._geom_names

    @property
    def geom_name2id(self):
        return self._geom_name2id

    @property
    def geom_id2name(self):
        return self._geom_id2name

    @property
    def site_names(self):
        return self._site_names

    @property
    def site_name2id(self):
        return self._site_name2id

    @property
    def site_id2name(self):
        return self._site_id2name

    @property
    def camera_names(self):
        return self._camera_names

    @property
    def camera_name2id(self):
        return self._camera_name2id

    @property
    def camera_id2name(self):
        return self._camera_id2name

    @property
    def actuator_names(self):
        return self._actuator_names

    @property
    def actuator_name2id(self):
        return self._actuator_name2id

    @property
    def actuator_id2name(self):
        return self._actuator_id2name

    @property
    def sensor_names(self):
        return self._sensor_names

    @property
    def sensor_name2id(self):
        return self._sensor_name2id

    @property
    def sensor_id2name(self):
        return self._sensor_id2name



#!/usr/bin/env python3
"""
测试CUDA和torch是否正确安装和启用的脚本
用于验证环境配置是否满足运行legged_gym的要求
"""

import sys
import subprocess
import platform

def check_python_version():
    """检查Python版本"""
    print("=" * 50)
    print("Python版本检查")
    print("=" * 50)
    print(f"Python版本: {sys.version}")
    print(f"Python路径: {sys.executable}")
    
    if sys.version_info < (3, 7):
        print("❌ Python版本过低，建议使用Python 3.7或更高版本")
        return False
    else:
        print("✅ Python版本符合要求")
        return True

def check_cuda_installation():
    """检查CUDA安装"""
    print("\n" + "=" * 50)
    print("CUDA安装检查")
    print("=" * 50)
    
    # 检查nvidia-smi命令
    try:
        result = subprocess.run(['nvidia-smi'], capture_output=True, text=True, timeout=10)
        if result.returncode == 0:
            print("✅ nvidia-smi命令可用")
            print("GPU信息:")
            print(result.stdout)
        else:
            print("❌ nvidia-smi命令不可用")
            return False
    except FileNotFoundError:
        print("❌ 未找到nvidia-smi命令，可能CUDA未安装或PATH中未包含")
        return False
    except subprocess.TimeoutExpired:
        print("❌ nvidia-smi命令执行超时")
        return False
    
    # 检查CUDA版本
    try:
        result = subprocess.run(['nvcc', '--version'], capture_output=True, text=True, timeout=10)
        if result.returncode == 0:
            print("✅ nvcc命令可用")
            # 提取CUDA版本
            for line in result.stdout.split('\n'):
                if 'release' in line.lower():
                    print(f"CUDA版本: {line.strip()}")
                    break
        else:
            print("❌ nvcc命令不可用")
            return False
    except FileNotFoundError:
        print("❌ 未找到nvcc命令，可能CUDA toolkit未安装或PATH中未包含")
        return False
    except subprocess.TimeoutExpired:
        print("❌ nvcc命令执行超时")
        return False
    
    return True

def check_torch_installation():
    """检查torch安装"""
    print("\n" + "=" * 50)
    print("PyTorch安装检查")
    print("=" * 50)
    
    try:
        import torch
        print(f"✅ PyTorch版本: {torch.__version__}")
        
        # 检查CUDA是否可用
        if torch.cuda.is_available():
            print("✅ PyTorch CUDA支持已启用")
            print(f"CUDA版本: {torch.version.cuda}")
            print(f"cuDNN版本: {torch.backends.cudnn.version()}")
            print(f"可用GPU数量: {torch.cuda.device_count()}")
            
            # 显示GPU信息
            for i in range(torch.cuda.device_count()):
                gpu_name = torch.cuda.get_device_name(i)
                gpu_memory = torch.cuda.get_device_properties(i).total_memory / 1024**3
                print(f"GPU {i}: {gpu_name} ({gpu_memory:.1f} GB)")
            
            # 测试CUDA张量操作
            try:
                x = torch.randn(1000, 1000).cuda()
                y = torch.randn(1000, 1000).cuda()
                z = torch.mm(x, y)
                print("✅ CUDA张量运算测试通过")
            except Exception as e:
                print(f"❌ CUDA张量运算测试失败: {e}")
                return False
                
        else:
            print("❌ PyTorch CUDA支持未启用")
            print("可能的原因:")
            print("1. 安装的是CPU版本的PyTorch")
            print("2. CUDA版本与PyTorch不匹配")
            print("3. 环境变量配置问题")
            return False
            
    except ImportError:
        print("❌ PyTorch未安装")
        return False
    except Exception as e:
        print(f"❌ PyTorch检查过程中出现错误: {e}")
        return False
    
    return True

def check_system_info():
    """检查系统信息"""
    print("\n" + "=" * 50)
    print("系统信息")
    print("=" * 50)
    print(f"操作系统: {platform.system()}")
    print(f"系统版本: {platform.release()}")
    print(f"架构: {platform.machine()}")
    print(f"处理器: {platform.processor()}")

def main():
    """主函数"""
    print("CUDA和PyTorch环境检查工具")
    print("用于验证legged_gym运行环境")
    
    # 检查系统信息
    check_system_info()
    
    # 检查Python版本
    python_ok = check_python_version()
    
    # 检查CUDA安装
    cuda_ok = check_cuda_installation()
    
    # 检查PyTorch安装
    torch_ok = check_torch_installation()
    
    # 总结
    print("\n" + "=" * 50)
    print("检查结果总结")
    print("=" * 50)
    
    if python_ok and cuda_ok and torch_ok:
        print("🎉 所有检查都通过了！环境配置正确，可以继续后续步骤。")
        return True
    else:
        print("⚠️  存在一些问题，请根据上述提示进行修复：")
        if not python_ok:
            print("- 升级Python版本")
        if not cuda_ok:
            print("- 安装或配置CUDA")
        if not torch_ok:
            print("- 重新安装匹配的PyTorch版本")
        return False

if __name__ == "__main__":
    success = main()
    sys.exit(0 if success else 1) 


import numpy as np
from gymnasium.core import ObsType
from orca_gym.utils import rotations
from typing import Optional, Any, SupportsFloat
from gymnasium import spaces
from orca_gym.environment.orca_gym_local_env import OrcaGymLocalEnv


class SimEnv(OrcaGymLocalEnv):
    """
    A class to represent the ORCA Gym environment for the Replicator scene.
    """

    def __init__(
        self,
        frame_skip: int,        
        orcagym_addr: str,
        agent_names: list,
        time_step: float,
        max_steps: Optional[int] = None,
        **kwargs,
    ):
        
        super().__init__(
            frame_skip = frame_skip,
            orcagym_addr = orcagym_addr,
            agent_names = agent_names,
            time_step = time_step,            
            **kwargs,
        )

        # Three auxiliary variables to understand the component of the xml document but will not be used
        # number of actuators/controls: 7 arm joints and 2 gripper joints
        self.nu = self.model.nu
        # 16 generalized coordinates: 9 (arm + gripper) + 7 (object free joint: 3 position and 4 quaternion coordinates)
        self.nq = self.model.nq
        # 9 arm joints and 6 free joints
        self.nv = self.model.nv

        self._set_obs_space()
        self._set_action_space()

    def _set_obs_space(self):
        self.observation_space = self.generate_observation_space(self._get_obs().copy())

    def _set_action_space(self):
        # 归一化到 [-1, 1]区间
        if (self.nu > 0):
            scaled_action_range = np.concatenate([[[-1.0, 1.0]] for _ in range(self.nu)])
            # print("Scaled action range: ", scaled_action_range)
            self.action_space = self.generate_action_space(scaled_action_range)
        else:
            self.action_space = spaces.Box(
                low=np.array([]),
                high=np.array([]),
                dtype=np.float32
            )
            print("No action space defined, nu is 0.")

    
    def render_callback(self, mode='human') -> None:
        if mode == "human":
            self.render()
        else:
            raise ValueError("Invalid render mode")

    def step(self, action) -> tuple:

        ctrl = np.zeros(self.nu, dtype=np.float32)

        # print("runmode: ", self._run_mode, "no_scaled_action: ", noscaled_action, "scaled_action: ", scaled_action, "ctrl: ", ctrl)
        
        # step the simulation with original action space
        self.do_simulation(ctrl, self.frame_skip)
        obs = self._get_obs().copy()

        info = {}
        terminated = False
        truncated = False
        reward = 0.0

        return obs, reward, terminated, truncated, info
    

    def _get_obs(self) -> dict:
           
        obs = {
            "joint_pos": self.data.qpos[:self.nq].copy(),
            "joint_vel": self.data.qvel[:self.nv].copy(),
            "joint_acc": self.data.qacc[:self.nv].copy(),
        }
        return obs


    def reset_model(self) -> tuple[dict, dict]:
        """
        Reset the environment, return observation
        """

        self.ctrl = np.zeros(self.nu, dtype=np.float32)

        obs = self._get_obs().copy()
        return obs, {}
    


    def get_observation(self, obs=None) -> dict:
        if obs is not None:
            return obs
        else:
            return self._get_obs().copy()



import numpy as np


def sensor(modality):
    """
    Decorator that should be added to any sensors that will be an observable.

    Decorated functions should have signature:

        any = func(obs_cache)

    Where @obs_cache is a dictionary mapping observable keys to pre-computed values, and @any is either a scalar
    or array. This function should also handle the case if obs_cache is either None or an empty dict.

    An example use case is shown below:

        >>> @sensor(modality="proprio")
        >>> def joint_pos(obs_cache):
                # Always handle case if obs_cache is empty
                if not obs_cache:
                    return np.zeros(7)
                # Otherwise, run necessary calculations and return output
                ...
                out = ...
                return out

    Args:
        modality (str): Modality for this sensor

    Returns:
        function: decorator function
    """
    # Define standard decorator (with no args)
    def decorator(func):
        # Add modality attribute
        func.__modality__ = modality
        # Return function
        return func

    return decorator


def create_deterministic_corrupter(corruption, low=-np.inf, high=np.inf):
    """
    Creates a deterministic corrupter that applies the same corrupted value to all sensor values

    Args:
        corruption (float): Corruption to apply
        low (float): Minimum value for output for clipping
        high (float): Maximum value for output for clipping

    Returns:
        function: corrupter
    """

    def corrupter(inp):
        inp = np.array(inp)
        return np.clip(inp + corruption, low, high)

    return corrupter


def create_uniform_noise_corrupter(min_noise, max_noise, low=-np.inf, high=np.inf):
    """
    Creates a corrupter that applies uniform noise to a given input within range @low to @high

    Args:
        min_noise (float): Minimum noise to apply
        max_noise (float): Maximum noise to apply
        low (float): Minimum value for output for clipping
        high (float): Maxmimum value for output for clipping

    Returns:
        function: corrupter
    """

    def corrupter(inp):
        inp = np.array(inp)
        noise = (max_noise - min_noise) * np.random.random_sample(inp.shape) + min_noise
        return np.clip(inp + noise, low, high)

    return corrupter


def create_gaussian_noise_corrupter(mean, std, low=-np.inf, high=np.inf):
    """
    Creates a corrupter that applies gaussian noise to a given input with mean @mean and std dev @std

    Args:
        mean (float): Mean of the noise to apply
        std (float): Standard deviation of the noise to apply
        low (float): Minimum value for output for clipping
        high (float): Maxmimum value for output for clipping

    Returns:
        function: corrupter
    """

    def corrupter(inp):
        inp = np.array(inp)
        noise = mean + std * np.random.randn(*inp.shape)
        return np.clip(inp + noise, low, high)

    return corrupter


def create_deterministic_delayer(delay):
    """
    Create a deterministic delayer that always returns the same delay value

    Args:
        delay (float): Delay value to return

    Returns:
        function: delayer
    """
    assert delay >= 0, "Inputted delay must be non-negative!"
    return lambda: delay


def create_uniform_sampled_delayer(min_delay, max_delay):
    """
    Creates uniformly sampled delayer, with minimum delay @low and maximum delay @high, both inclusive

    Args:
        min_delay (float): Minimum possible delay
        max_delay (float): Maxmimum possible delay

    Returns:
        function: delayer
    """
    assert min(min_delay, max_delay) >= 0, "Inputted delay must be non-negative!"
    return lambda: min_delay + (max_delay - min_delay) * np.random.random()


def create_gaussian_sampled_delayer(mean, std):
    """
    Creates a gaussian sampled delayer, with average delay @mean which varies by standard deviation @std

    Args:
        mean (float): Average delay
        std (float): Standard deviation of the delay variation

    Returns:
        function: delayer
    """
    assert mean >= 0, "Inputted mean delay must be non-negative!"
    return lambda: max(0.0, int(np.round(mean + std * np.random.randn())))


# Common defaults to use
NO_CORRUPTION = lambda inp: inp
NO_FILTER = lambda inp: inp
NO_DELAY = lambda: 0.0


class Observable:
    """
    Base class for all observables -- defines interface for interacting with sensors

    Args:
        name (str): Name for this observable
        sensor (function with `sensor` decorator): Method to grab raw sensor data for this observable. Should take in a
            single dict argument (observation cache if a pre-computed value is required) and return the raw sensor data
            for the current timestep. Must handle case if inputted argument is empty ({}), and should have `sensor`
            decorator when defined
        corrupter (None or function): Method to corrupt the raw sensor data for this observable. Should take in
            the output of @sensor and return the same type (corrupted data). If None, results in default no corruption
        filter (None or function): Method to filter the outputted reading for this observable. Should take in the output
            of @corrupter and return the same type (filtered data). If None, results in default no filter. Note that
            this function can also double as an observer, where sampled data is recorded by this function.
        delayer (None or function): Method to delay the raw sensor data when polling this observable. Should take in
            no arguments and return a float, for the number of seconds to delay the measurement by. If None, results in
            default no delayer
        sampling_rate (float): Sampling rate for this observable (Hz)
        enabled (bool): Whether this sensor is enabled or not. If enabled, this observable's values
            are continually computed / updated every time update() is called.
        active (bool): Whether this sensor is active or not. If active, this observable's current
            observed value is returned from self.obs, otherwise self.obs returns None.
    """

    def __init__(
        self,
        name,
        sensor,
        corrupter=None,
        filter=None,
        delayer=None,
        sampling_rate=20,
        enabled=True,
        active=True,
    ):
        # Set all internal variables and methods
        self.name = name
        self._sensor = sensor
        self._corrupter = corrupter if corrupter is not None else NO_CORRUPTION
        self._filter = filter if filter is not None else NO_FILTER
        self._delayer = delayer if delayer is not None else NO_DELAY
        self._sampling_timestep = 1.0 / sampling_rate
        self._enabled = enabled
        self._active = active
        self._is_number = False  # filled in during sensor check call
        self._data_shape = (1,)  # filled in during sensor check call

        # Make sure sensor is working
        self._check_sensor_validity()

        # These values will be modified during update() call
        self._time_since_last_sample = 0.0  # seconds
        self._current_delay = self._delayer()  # seconds
        self._current_observed_value = 0 if self._is_number else np.zeros(self._data_shape)
        self._sampled = False

    def update(self, timestep, obs_cache, force=False):
        """
        Updates internal values for this observable, if enabled.

        Args:
            timestep (float): Amount of simulation time (in sec) that has passed since last call.
            obs_cache (dict): Observation cache mapping observable names to pre-computed values to pass to sensor. This
                will be updated in-place during this call.
            force (bool): If True, will force the observable to update its internal value to the newest value.
        """
        if self._enabled:
            # Increment internal time counter
            self._time_since_last_sample += timestep

            # If the delayed sampling time has been passed and we haven't sampled yet for this sampling period,
            # we should grab a new measurement
            if (
                not self._sampled and self._sampling_timestep - self._current_delay >= self._time_since_last_sample
            ) or force:
                # Get newest raw value, corrupt it, filter it, and set it as our current observed value
                obs = np.array(self._filter(self._corrupter(self._sensor(obs_cache))))
                self._current_observed_value = obs[0] if len(obs.shape) == 1 and obs.shape[0] == 1 else obs
                # Update cache entry as well
                obs_cache[self.name] = np.array(self._current_observed_value)
                # Toggle sampled and re-sample next time delay
                self._sampled = True
                self._current_delay = self._delayer()

            # If our total time since last sample has surpassed our sampling timestep,
            # then we reset our timer and sampled flag
            if self._time_since_last_sample >= self._sampling_timestep:
                if not self._sampled:
                    # If we still haven't sampled yet, sample immediately and warn user that sampling rate is too low
                    print(
                        f"Warning: sampling rate for observable {self.name} is either too low or delay is too high. "
                        f"Please adjust one (or both)"
                    )
                    # Get newest raw value, corrupt it, filter it, and set it as our current observed value
                    obs = np.array(self._filter(self._corrupter(self._sensor(obs_cache))))
                    self._current_observed_value = obs[0] if len(obs.shape) == 1 and obs.shape[0] == 1 else obs
                    # Update cache entry as well
                    obs_cache[self.name] = np.array(self._current_observed_value)
                    # Re-sample next time delay
                    self._current_delay = self._delayer()
                self._time_since_last_sample %= self._sampling_timestep
                self._sampled = False

    def reset(self):
        """
        Resets this observable's internal values (but does not reset its sensor, corrupter, delayer, or filter)
        """
        self._time_since_last_sample = 0.0
        self._current_delay = self._delayer()
        self._current_observed_value = 0 if self._is_number else np.zeros(self._data_shape)

    def is_enabled(self):
        """
        Determines whether observable is enabled or not. This observable is considered enabled if its values
        are being continually computed / updated during each update() call.

        Returns:
            bool: True if this observable is enabled
        """
        return self._enabled

    def is_active(self):
        """
        Determines whether observable is active or not. This observable is considered active if its current observation
        value is being returned in self.obs.

        Returns:
            bool: True if this observable is active
        """
        return self._active

    def set_enabled(self, enabled):
        """
        Sets whether this observable is enabled or not. If enabled, this observable's values
        are continually computed / updated every time update() is called.

        Args:
            enabled (bool): True if this observable should be enabled
        """
        self._enabled = enabled
        # Reset values
        self.reset()

    def set_active(self, active):
        """
        Sets whether this observable is active or not. If active, this observable's current
        observed value is returned from self.obs, otherwise self.obs returns None.

        Args:
            active (bool): True if this observable should be active
        """
        self._active = active

    def set_sensor(self, sensor):
        """
        Sets the sensor for this observable.

        Args:
            sensor (function with sensor decorator): Method to grab raw sensor data for this observable. Should take in
                a single dict argument (observation cache if a pre-computed value is required) and return the raw
                sensor data for the current timestep. Must handle case if inputted argument is empty ({}), and should
                have `sensor` decorator when defined
        """
        self._sensor = sensor
        self._check_sensor_validity()

    def set_corrupter(self, corrupter):
        """
        Sets the corrupter for this observable.

        Args:
             corrupter (None or function): Method to corrupt the raw sensor data for this observable. Should take in
                the output of self.sensor and return the same type (corrupted data).
                If None, results in default no corruption
        """
        self._corrupter = corrupter if corrupter is not None else NO_CORRUPTION

    def set_filter(self, filter):
        """
        Sets the filter for this observable. Note that this function can also double as an observer, where sampled
        data is recorded by this function.

        Args:
             filter (None or function): Method to filter the outputted reading for this observable. Should take in
                the output of @corrupter and return the same type (filtered data).
                If None, results in default no filter
        """
        self._filter = filter if filter is not None else NO_FILTER

    def set_delayer(self, delayer):
        """
        Sets the delayer for this observable.

        Args:
            delayer (None or function): Method to delay the raw sensor data when polling this observable. Should take
                in no arguments and return a float, for the number of seconds to delay the measurement by.
                If None, results in default no filter
        """
        self._delayer = delayer if delayer is not None else NO_DELAY

    def set_sampling_rate(self, rate):
        """
        Sets the sampling rate for this observable.

        Args:
            rate (int): New sampling rate for this observable (Hz)
        """
        self._sampling_timestep = 1.0 / rate

    def _check_sensor_validity(self):
        """
        Internal function that checks the validity of this observable's sensor. It does the following:

            - Asserts that the inputted sensor has its __modality__ attribute defined from the sensor decorator
            - Asserts that the inputted sensor can handle the empty dict {} arg case
            - Updates the corresponding name, and data-types for this sensor
        """
        try:
            _ = self.modality
            self._data_shape = np.array(self._sensor({})).shape
            self._is_number = len(self._data_shape) == 1 and self._data_shape[0] == 1
        except Exception as e:
            from orca_gym.adapters.robosuite.utils.log_utils import ROBOSUITE_DEFAULT_LOGGER

            ROBOSUITE_DEFAULT_LOGGER.error(e)
            raise ValueError("Current sensor for observable {} is invalid.".format(self.name))

    @property
    def obs(self):
        """
        Current observation from this observable

        Returns:
            None or float or np.array: If active, current observed value from this observable. Otherwise, None
        """
        return self._current_observed_value if self._active else None

    @property
    def modality(self):
        """
        Modality of this sensor

        Returns:
            str: Modality name for this observable
        """
        return self._sensor.__modality__



import abc
from collections.abc import Iterable

from orca_gym import OrcaGymLocal
import numpy as np

import orca_gym.adapters.robosuite.macros as macros
import asyncio


class Controller(object, metaclass=abc.ABCMeta):
    """
    General controller interface.

    Requires reference to mujoco sim object, eef_name of specific robot, relevant joint_indexes to that robot, and
    whether an initial_joint is used for nullspace torques or not

    Args:
        sim (MjSim): Simulator instance this controller will pull robot state updates from

        eef_name (str): Name of controlled robot arm's end effector (from robot XML)

        joint_indexes (dict): Each key contains sim reference indexes to relevant robot joint information, namely:

            :`'joints'`: list of indexes to relevant robot joints
            :`'qpos'`: list of indexes to relevant robot joint positions
            :`'qvel'`: list of indexes to relevant robot joint velocities

        actuator_range (2-tuple of array of float): 2-Tuple (low, high) representing the robot joint actuator range
    """

    def __init__(
        self,
        gym: OrcaGymLocal,
        eef_name,
        joint_indexes,
        actuator_range,
    ):
        # Actuator range
        self.actuator_min = actuator_range[0]
        self.actuator_max = actuator_range[1]

        # Attributes for scaling / clipping inputs to outputs
        self.action_scale = None
        self.action_input_transform = None
        self.action_output_transform = None

        # Private property attributes
        self.control_dim = None
        self.output_min = None
        self.output_max = None
        self.input_min = None
        self.input_max = None

        # mujoco simulator state
        self.gym = gym
        self.model_timestep = gym.opt.timestep   #macros.SIMULATION_TIMESTEP
        self.eef_name = eef_name
        self.joint_index = joint_indexes["joints"]
        self.qpos_index = joint_indexes["qpos"]
        self.qvel_index = joint_indexes["qvel"]

        # robot states
        self.ee_pos = None
        self.ee_ori_mat = None
        self.ee_pos_vel = None
        self.ee_ori_vel = None
        self.joint_pos = None
        self.joint_vel = None

        # dynamics and kinematics
        self.J_pos = None
        self.J_ori = None
        self.J_full = None
        self.mass_matrix = None

        # Joint dimension
        self.joint_dim = len(joint_indexes["joints"])

        # Torques being outputted by the controller
        self.torques = None

        # Update flag to prevent redundant update calls
        self.new_update = True

        # Move forward one timestep to propagate updates before taking first update
        self.gym.mj_forward()

        # Initialize controller by updating internal state and setting the initial joint, pos, and ori
        self.update()
        self.initial_joint = self.joint_pos
        self.initial_ee_pos = self.ee_pos
        self.initial_ee_ori_mat = self.ee_ori_mat

    @abc.abstractmethod
    def run_controller(self):
        """
        Abstract method that should be implemented in all subclass controllers, and should convert a given action
        into torques (pre gravity compensation) to be executed on the robot.
        Additionally, resets the self.new_update flag so that the next self.update call will occur
        """
        self.new_update = True

    def scale_action(self, action):
        """
        Clips @action to be within self.input_min and self.input_max, and then re-scale the values to be within
        the range self.output_min and self.output_max

        Args:
            action (Iterable): Actions to scale

        Returns:
            np.array: Re-scaled action
        """

        if self.action_scale is None:
            self.action_scale = abs(self.output_max - self.output_min) / abs(self.input_max - self.input_min)
            self.action_output_transform = (self.output_max + self.output_min) / 2.0
            self.action_input_transform = (self.input_max + self.input_min) / 2.0
        action = np.clip(action, self.input_min, self.input_max)
        transformed_action = (action - self.action_input_transform) * self.action_scale + self.action_output_transform

        return transformed_action

    def update(self, force=False):
        """
        Updates the state of the robot arm, including end effector pose / orientation / velocity, joint pos/vel,
        jacobian, and mass matrix. By default, since this is a non-negligible computation, multiple redundant calls
        will be ignored via the self.new_update attribute flag. However, if the @force flag is set, the update will
        occur regardless of that state of self.new_update. This base class method of @run_controller resets the
        self.new_update flag

        Args:
            force (bool): Whether to force an update to occur or not
        """

        # Only run update if self.new_update or force flag is set
        if self.new_update or force:

            ee_pos_mat_dict = self.gym.query_site_pos_and_mat([self.eef_name])
            jac_site_dict = self.gym.mj_jac_site([self.eef_name])
            mass_matrix = self.gym.mj_fullM()
            # print("mass_matrix: ", mass_matrix)
                
            self.ee_pos = np.array(ee_pos_mat_dict[self.eef_name]['xpos'])
            self.ee_ori_mat = np.array(ee_pos_mat_dict[self.eef_name]['xmat']).reshape((3, 3))

            # self.ee_pos = np.array(self.sim.data.site_xpos[self.sim.model.site_name2id(self.eef_name)])
            # self.ee_ori_mat = np.array(
            #     self.sim.data.site_xmat[self.sim.model.site_name2id(self.eef_name)].reshape([3, 3])
            # )

            jacp = jac_site_dict[self.eef_name]['jacp'].reshape((3, -1))
            jacr = jac_site_dict[self.eef_name]['jacr'].reshape((3, -1))
            self.ee_pos_vel = np.dot(jacp, self.gym.data.qvel)
            self.ee_ori_vel = np.dot(jacr, self.gym.data.qvel)

            # self.ee_pos_vel = np.array(self.sim.data.get_site_xvelp(self.eef_name))
            # self.ee_ori_vel = np.array(self.sim.data.get_site_xvelr(self.eef_name))

            self.joint_pos = np.array(self.gym.data.qpos[self.qpos_index])
            self.joint_vel = np.array(self.gym.data.qvel[self.qvel_index])

            self.J_pos = np.array(jacp[:, self.qvel_index])
            self.J_ori = np.array(jacr[:, self.qvel_index])

            # self.J_pos = np.array(self.sim.data.get_site_jacp(self.eef_name).reshape((3, -1))[:, self.qvel_index])
            # self.J_ori = np.array(self.sim.data.get_site_jacr(self.eef_name).reshape((3, -1))[:, self.qvel_index])

            self.J_full = np.array(np.vstack([self.J_pos, self.J_ori]))

            # mass_matrix = np.ndarray(shape=(self.sim.model.nv, self.sim.model.nv), dtype=np.float64, order="C")
            # mujoco.mj_fullM(self.sim.model._model, mass_matrix, self.sim.data.qM)
            # mass_matrix = np.reshape(mass_matrix, (len(self.sim.data.qvel), len(self.sim.data.qvel)))

            self.mass_matrix = mass_matrix[self.qvel_index, :][:, self.qvel_index]

            # Clear self.new_update
            self.new_update = False

    def update_base_pose(self, base_pos, base_ori):
        """
        Optional function to implement in subclass controllers that will take in @base_pos and @base_ori and update
        internal configuration to account for changes in the respective states. Useful for controllers e.g. IK, which
        is based on pybullet and requires knowledge of simulator state deviations between pybullet and mujoco

        Args:
            base_pos (3-tuple): x,y,z position of robot base in mujoco world coordinates
            base_ori (4-tuple): x,y,z,w orientation or robot base in mujoco world coordinates
        """
        pass

    def update_initial_joints(self, initial_joints):
        """
        Updates the internal attribute self.initial_joints. This is useful for updating changes in controller-specific
        behavior, such as with OSC where self.initial_joints is used for determine nullspace actions

        This function can also be extended by subclassed controllers for additional controller-specific updates

        Args:
            initial_joints (Iterable): Array of joint position values to update the initial joints
        """
        self.initial_joint = np.array(initial_joints)
        self.update(force=True)
        self.initial_ee_pos = self.ee_pos
        self.initial_ee_ori_mat = self.ee_ori_mat

    def clip_torques(self, torques):
        """
        Clips the torques to be within the actuator limits

        Args:
            torques (Iterable): Torques to clip

        Returns:
            np.array: Clipped torques
        """
        return np.clip(torques, self.actuator_min, self.actuator_max)

    def reset_goal(self):
        """
        Resets the goal -- usually by setting to the goal to all zeros, but in some cases may be different (e.g.: OSC)
        """
        raise NotImplementedError

    @staticmethod
    def nums2array(nums, dim):
        """
        Convert input @nums into numpy array of length @dim. If @nums is a single number, broadcasts it to the
        corresponding dimension size @dim before converting into a numpy array

        Args:
            nums (numeric or Iterable): Either single value or array of numbers
            dim (int): Size of array to broadcast input to env.sim.data.actuator_force

        Returns:
            np.array: Array filled with values specified in @nums
        """
        # First run sanity check to make sure no strings are being inputted
        if isinstance(nums, str):
            raise TypeError("Error: Only numeric inputs are supported for this function, nums2array!")

        # Check if input is an Iterable, if so, we simply convert the input to np.array and return
        # Else, input is a single value, so we map to a numpy array of correct size and return
        return np.array(nums) if isinstance(nums, Iterable) else np.ones(dim) * nums

    @property
    def torque_compensation(self):
        """
        Gravity compensation for this robot arm

        Returns:
            np.array: torques
        """
        return self.gym.data.qfrc_bias[self.qvel_index]

    @property
    def actuator_limits(self):
        """
        Torque limits for this controller

        Returns:
            2-tuple:

                - (np.array) minimum actuator torques
                - (np.array) maximum actuator torques
        """
        return self.actuator_min, self.actuator_max

    @property
    def control_limits(self):
        """
        Limits over this controller's action space, which defaults to input min/max

        Returns:
            2-tuple:

                - (np.array) minimum action values
                - (np.array) maximum action values
        """
        return self.input_min, self.input_max

    @property
    def name(self):
        """
        Name of this controller

        Returns:
            str: controller name
        """
        raise NotImplementedError



"""
Utility functions for grabbing user inputs
"""

import numpy as np

import orca_gym.adapters.robosuite.as suite
import orca_gym.adapters.robosuite.utils.transform_utils as T
from orca_gym.adapters.robosuite.devices import *
from orca_gym.adapters.robosuite.models.robots import *
from orca_gym.adapters.robosuite.robots import *


def choose_environment():
    """
    Prints out environment options, and returns the selected env_name choice

    Returns:
        str: Chosen environment name
    """
    # get the list of all environments
    envs = sorted(suite.ALL_ENVIRONMENTS)

    # Select environment to run
    print("Here is a list of environments in the suite:\n")

    for k, env in enumerate(envs):
        print("[{}] {}".format(k, env))
    print()
    try:
        s = input("Choose an environment to run " + "(enter a number from 0 to {}): ".format(len(envs) - 1))
        # parse input into a number within range
        k = min(max(int(s), 0), len(envs))
    except:
        k = 0
        print("Input is not valid. Use {} by default.\n".format(envs[k]))

    # Return the chosen environment name
    return envs[k]


def choose_controller():
    """
    Prints out controller options, and returns the requested controller name

    Returns:
        str: Chosen controller name
    """
    # get the list of all controllers
    controllers_info = suite.controllers.CONTROLLER_INFO
    controllers = list(suite.ALL_CONTROLLERS)

    # Select controller to use
    print("Here is a list of controllers in the suite:\n")

    for k, controller in enumerate(controllers):
        print("[{}] {} - {}".format(k, controller, controllers_info[controller]))
    print()
    try:
        s = input("Choose a controller for the robot " + "(enter a number from 0 to {}): ".format(len(controllers) - 1))
        # parse input into a number within range
        k = min(max(int(s), 0), len(controllers) - 1)
    except:
        k = 0
        print("Input is not valid. Use {} by default.".format(controllers)[k])

    # Return chosen controller
    return controllers[k]


def choose_multi_arm_config():
    """
    Prints out multi-arm environment configuration options, and returns the requested config name

    Returns:
        str: Requested multi-arm configuration name
    """
    # Get the list of all multi arm configs
    env_configs = {
        "Single Arms Opposed": "single-arm-opposed",
        "Single Arms Parallel": "single-arm-parallel",
        "Bimanual": "bimanual",
    }

    # Select environment configuration
    print("A multi-arm environment was chosen. Here is a list of multi-arm environment configurations:\n")

    for k, env_config in enumerate(list(env_configs)):
        print("[{}] {}".format(k, env_config))
    print()
    try:
        s = input(
            "Choose a configuration for this environment "
            + "(enter a number from 0 to {}): ".format(len(env_configs) - 1)
        )
        # parse input into a number within range
        k = min(max(int(s), 0), len(env_configs))
    except:
        k = 0
        print("Input is not valid. Use {} by default.".format(list(env_configs)[k]))

    # Return requested configuration
    return list(env_configs.values())[k]


def choose_robots(exclude_bimanual=False):
    """
    Prints out robot options, and returns the requested robot. Restricts options to single-armed robots if
    @exclude_bimanual is set to True (False by default)

    Args:
        exclude_bimanual (bool): If set, excludes bimanual robots from the robot options

    Returns:
        str: Requested robot name
    """
    # Get the list of robots
    robots = {
        "Sawyer",
        "Panda",
        "Jaco",
        "Kinova3",
        "IIWA",
        "UR5e",
    }

    # Add Baxter if bimanual robots are not excluded
    if not exclude_bimanual:
        robots.add("Baxter")

    # Make sure set is deterministically sorted
    robots = sorted(robots)

    # Select robot
    print("Here is a list of available robots:\n")

    for k, robot in enumerate(robots):
        print("[{}] {}".format(k, robot))
    print()
    try:
        s = input("Choose a robot " + "(enter a number from 0 to {}): ".format(len(robots) - 1))
        # parse input into a number within range
        k = min(max(int(s), 0), len(robots))
    except:
        k = 0
        print("Input is not valid. Use {} by default.".format(list(robots)[k]))

    # Return requested robot
    return list(robots)[k]


def input2action(device, robot, active_arm="right", env_configuration=None):
    """
    Converts an input from an active device into a valid action sequence that can be fed into an env.step() call

    If a reset is triggered from the device, immediately returns None. Else, returns the appropriate action

    Args:
        device (Device): A device from which user inputs can be converted into actions. Can be either a Spacemouse or
            Keyboard device class

        robot (Robot): Which robot we're controlling

        active_arm (str): Only applicable for multi-armed setups (e.g.: multi-arm environments or bimanual robots).
            Allows inputs to be converted correctly if the control type (e.g.: IK) is dependent on arm choice.
            Choices are {right, left}

        env_configuration (str or None): Only applicable for multi-armed environments. Allows inputs to be converted
            correctly if the control type (e.g.: IK) is dependent on the environment setup. Options are:
            {bimanual, single-arm-parallel, single-arm-opposed}

    Returns:
        2-tuple:

            - (None or np.array): Action interpreted from @device including any gripper action(s). None if we get a
                reset signal from the device
            - (None or int): 1 if desired close, -1 if desired open gripper state. None if get a reset signal from the
                device

    """
    state = device.get_controller_state()
    # Note: Devices output rotation with x and z flipped to account for robots starting with gripper facing down
    #       Also note that the outputted rotation is an absolute rotation, while outputted dpos is delta pos
    #       Raw delta rotations from neutral user input is captured in raw_drotation (roll, pitch, yaw)
    dpos, rotation, raw_drotation, grasp, reset = (
        state["dpos"],
        state["rotation"],
        state["raw_drotation"],
        state["grasp"],
        state["reset"],
    )

    # If we're resetting, immediately return None
    if reset:
        return None, None

    # Get controller reference
    controller = robot.controller if not isinstance(robot, Bimanual) else robot.controller[active_arm]
    gripper_dof = robot.gripper.dof if not isinstance(robot, Bimanual) else robot.gripper[active_arm].dof

    # First process the raw drotation
    drotation = raw_drotation[[1, 0, 2]]
    if controller.name == "IK_POSE":
        # If this is panda, want to swap x and y axis
        if isinstance(robot.robot_model, Panda):
            drotation = drotation[[1, 0, 2]]
        else:
            # Flip x
            drotation[0] = -drotation[0]
        # Scale rotation for teleoperation (tuned for IK)
        drotation *= 10
        dpos *= 5
        # relative rotation of desired from current eef orientation
        # map to quat
        drotation = T.mat2quat(T.euler2mat(drotation))

        # If we're using a non-forward facing configuration, need to adjust relative position / orientation
        if env_configuration == "single-arm-opposed":
            # Swap x and y for pos and flip x,y signs for ori
            dpos = dpos[[1, 0, 2]]
            drotation[0] = -drotation[0]
            drotation[1] = -drotation[1]
            if active_arm == "left":
                # x pos needs to be flipped
                dpos[0] = -dpos[0]
            else:
                # y pos needs to be flipped
                dpos[1] = -dpos[1]

        # Lastly, map to axis angle form
        drotation = T.quat2axisangle(drotation)

    elif controller.name == "OSC_POSE":
        # Flip z
        drotation[2] = -drotation[2]
        # Scale rotation for teleoperation (tuned for OSC) -- gains tuned for each device
        drotation = drotation * 1.5 if isinstance(device, Keyboard) else drotation * 50
        dpos = dpos * 75 if isinstance(device, Keyboard) else dpos * 125
    elif controller.name == "OSC_POSITION":
        dpos = dpos * 75 if isinstance(device, Keyboard) else dpos * 125
    else:
        # No other controllers currently supported
        print("Error: Unsupported controller specified -- Robot must have either an IK or OSC-based controller!")

    # map 0 to -1 (open) and map 1 to 1 (closed)
    grasp = 1 if grasp else -1

    # Create action based on action space of individual robot
    if controller.name == "OSC_POSITION":
        action = np.concatenate([dpos, [grasp] * gripper_dof])
    else:
        action = np.concatenate([dpos, drotation, [grasp] * gripper_dof])

    # Return the action and grasp
    return action, grasp



import math
import os
import shutil
import sys
import time
import subprocess
import signal
from filelock import FileLock, Timeout

from typing import Any, Dict
import uuid
import json
import gymnasium as gym
from gymnasium.envs.registration import register
from datetime import datetime, timedelta, timezone

import h5py
# from orca_gym.utils.kps_data_checker import BasicUnitChecker, ErrorType
from orca_gym.environment.orca_gym_env import RewardType
from orca_gym.adapters.robomimic.dataset_util import DatasetWriter, DatasetReader
from orca_gym.sensor.rgbd_camera import Monitor, CameraWrapper
from envs.manipulation.dual_arm_env import DualArmEnv, ControlDevice
from examples.imitation.train_policy import train_policy
from examples.imitation.test_policy import create_env, rollout
from orca_gym.utils.dir_utils import create_tmp_dir
from robomimic.utils.file_utils import maybe_dict_from_checkpoint
from robomimic.utils.train_utils import run_rollout
import orca_gym.utils.rotations as rotations
from envs.manipulation.dual_arm_env import TaskStatus
from envs.manipulation.dual_arm_env import ActionType
import re
import gymnasium as gym
from envs.manipulation.dual_arm_env import ControlDevice, RunMode, DualArmEnv
from envs.manipulation.dual_arm_robot import DualArmRobot
import numpy as np
import yaml

ENV_NAME = "DualArmEnv"
ENV_ENTRY_POINT = {
    "DualArmEnv": "envs.manipulation.dual_arm_env:DualArmEnv"
}

TIME_STEP = 0.001                       # 1000 Hz for physics simulation
FRAME_SKIP = 20                         
REALTIME_STEP = TIME_STEP * FRAME_SKIP  # 50 Hz for python program loop
CONTROL_FREQ = 1 / REALTIME_STEP        # 50 Hz for OSC controller computation

RGB_SIZE = (1280, 720)
CAMERA_CONFIG = {
    "camera_head": 7070,
    # "camera_wrist_r": 7080,
    # "camera_wrist_l": 7090,
}
_light_counter = 0
_LIGHT_SWITCH_PERIOD = 20  # 每 20 次 reset 才切一次光
INIT_SCENE_TEXT = {
    "shop":    ("机器人站在收银台前，1-5个可抓取物体随机分布在收银台上，扫码枪位于机器人右手边。",  "The robot stands in front of the cash register, 1-5 grabbable objects are randomly distributed on the cash register, and the code scanner is located on the right hand side of the robot."),
    "yaodian": ("一个机器人站在药柜前",  "A robot stands in front of a medicine cabinet."),
    "kitchen": ("一个机器人站在灶台前",  "A robot stands in front of a stove."),
    "jiazi":   ("一个机器人站在货架前",  "A robot stands in front of a shelf."),
    "pharmacy":   ("机器人站在阴凉柜前，蓝色的框子位于机器人前，1-5个可抓取药盒随机分布在阴凉柜架子上。",  "The robot stands in front of the cooler, and the blue box is located in front of the robot，and 1-5 grabbable pill boxes are randomly distributed on the cooler shelves."),
    "housekeeping":   ("机器人站在双开门冰箱前，冰箱门状态为以下四种状态（左开右闭、左闭右开、双门全闭、双门全开）之一。",  "The robot stands in front of the double-door refrigerator, and the refrigerator door state is one of the following four states (left open & right closed, left closed & right open, both doors closed, both doors open)."),
    "3c_fabrication": ("机器人站在操作台前，1-5个可抓取物体随机分布在黄色框子中，扫码枪位于机器人右手边。",  "The robot stands in front of the console, 1-5 grabbable objects are randomly distributed on the yellow box, and the code scanner is located on the right side of the robot."),
}
_light_counter = 0

g_skip_frame = 0

# 顶部：维护一个简单的中英映射


OBJ_CN = {
    "can": "易拉罐",
    "bottle_red":"红色瓶子",
    "bottle_blue":"蓝色瓶子",
    "jar_01": "罐子",
    "salt": "盐罐",
    "basket": "篮子",
    "medicinechest": "药柜",
    "niuhuangwan": "牛黄丸",
    "qinghuopian": "清火片",
    "orangebox" : "橙色药盒",
    "pipalu": "枇杷露",
    "xiaokepian": "消咳片",
    "bow_yellow_kps": "黄色碗",
    "changyanning": "肠炎宁",
    "fenghanganmao": "风寒感冒颗粒",
    "yiyakangpian": "益压康片",
    "xiaozhongzhitong": "消肿止痛酊",
    "coffeecup_white_kps": "白色咖啡杯",
    "basket_kitchen_01" : "篮子",
    "shoppingtrolley_01" : "购物手推车",
    "qianglipipalu": "强力枇杷露",
    "box_blue" : "蓝色筐子",
    "fridge_right_up":"冰箱右边",
    "fridge_left_up":"冰箱左边",
    "shop": "超市",
    "kitchen": "厨房",
    "yaodian": "药店",
    "3c_fabrication": "3c制造",
    "battery_01": "电池",
    "intel_box":"英特尔处理器",
    "intelcore_i5":"英特尔i5",
    "wifi_box":"路由器",
    "cpu_fan":"cpu风扇",
    "fridge":"冰箱",
    "housekeeping":"家政",
    "Guizi": "柜子",
    "guizi": "柜子",
    "clinic": "诊所",
    "pharmacy": "药店",
    "barcode": "扫码枪",
}
SCENE_SUBSCENE_MAPPING = {
    "shop":    ("Shop",    "Cashier_Operation"),
    "jiazi":   ("Shop",    "Shelf_Operation"),
    "kitchen": ("Kitchen", "Countertop_Operation"),
    "yaodian": ("Pharmacy","Shelf_Operation"),
    # "guizi": ("Cooler","Shelf_Operation")
    "pharmacy": ("pharmacy","Cooler_Operation"),
    "housekeeping": ("fridge","Fridge_Operation"),
    "3c_fabrication": ("3c_scan","3C_Scan_Operation")
}

with open("camera_config.yaml", "r") as f:
    cam_cfg_all = yaml.safe_load(f)
default_cfg = cam_cfg_all["default"]
scene_cfgs = cam_cfg_all["scenes"]

# —— 生成对应场景下的 CAMERA_STATIC_CFG ——
def get_camera_static_cfg(level_name: str):
    scene = scene_cfgs.get(level_name)
    if scene is None:
        raise KeyError(f"No camera config for scene '{level_name}'")
    cfg = {}
    for cam_name, base in default_cfg.items():
        extr = scene.get(cam_name)
        if extr is None:
            raise KeyError(f"No scene extrinsics for camera '{cam_name}' in scene '{level_name}'")
        cfg[cam_name] = {
            **base,
            "translation": base["translation"],       # 原始外参（机器人坐标系）
            "rotation":    base["rotation"],          # 原始外参（机器人坐标系）
            "translation_world": extr["translation"], # 对齐外参（世界坐标系）
            "rotation_world":    extr["rotation"],    # 对齐外参（世界坐标系）
        }
    return cfg

# 3) dump_static_camera_params，写出三种 JSON

def dump_static_camera_params(uuid_dir: str,
                              level_name: str
                              ):
    CAMERA_STATIC_CFG = get_camera_static_cfg(level_name)
    out_dir = os.path.join(uuid_dir, "parameters")
    os.makedirs(out_dir, exist_ok=True)

    for name, cfg in CAMERA_STATIC_CFG.items():
        # —— 内参 ——
        intrinsic = {
            "manufacturer": "openverse",
            "mode":         "Mujoco Camera Capture",
            "SN":           "CA85BDD5-C631-4C52-853C-06D654AE7E4D",
            "fps":          30,
            "width":        cfg["width"],
            "height":       cfg["height"],
            "intrinsic": {
                "fx": (cfg["height"]/2) / math.tan(math.radians(cfg["fovy"]/2)),
                "fy": (cfg["height"]/2) / math.tan(math.radians(cfg["fovy"]/2)),
                "ppx": cfg["width"]/2,
                "ppy": cfg["height"]/2,
                "distortion_model": "plumb_bob",
                "k1": 0, "k2": 0, "k3": 0, "p1": 0, "p2": 0
            }
        }
        # —— 原始 extrinsic（机器人坐标系） ——
        extrinsic = {
            "translation_vector": cfg["translation"],
            "rotation_euler":     cfg["rotation"],
        }
        # —— 对齐 extrinsic（世界坐标系） ——
        extrinsic_aligned = {
            "translation_vector": cfg["translation_world"],
            "rotation_euler":     cfg["rotation_world"],
        }

        # 写文件
        with open(os.path.join(out_dir, f"{name}_intrinsic_params.json"),   "w", encoding="utf-8") as f:
            json.dump(intrinsic, f, ensure_ascii=False, indent=2)
        with open(os.path.join(out_dir, f"{name}_extrinsic_params.json"),   "w", encoding="utf-8") as f:
            json.dump(extrinsic, f, ensure_ascii=False, indent=2)
        with open(os.path.join(out_dir, f"{name}_extrinsic_params_aligned.json"), "w", encoding="utf-8") as f:
            json.dump(extrinsic_aligned, f, ensure_ascii=False, indent=2)


def normalize_key(key: str) -> str:
    """去掉多余标点，统一小写"""
    return re.sub(r'[^\w]', '', key).lower()

def eng2cn(instruction_en: str,level_name: str = "") -> str:
    """
    把类似 "level: shop object: bottle_blue to goal: basket" 或
    "put bottle_blue into basket" 之类的英文指令，翻成中文。
    """
    text = instruction_en.strip().lower()
    # 如果没传 level_name，就从 instruction_en 里提取
    if not level_name:
        m_lvl = re.search(r'in level\s+([\w_]+)', text)
        if m_lvl:
            level_name = m_lvl.group(1)
    
    # 1) 先试"object … to goal …" 的结构
    m = re.search(r'object[:\s]+([\w_]+)\s+to\s+goal[:\s]+([\w_]+)', text)
    if m:
        obj_key  = normalize_key(m.group(1))
        goal_key = normalize_key(m.group(2))
        obj_cn   = OBJ_CN.get(obj_key, obj_key)
        goal_cn  = OBJ_CN.get(goal_key, goal_key)
        return f"将{obj_cn}放入{goal_cn}中"
    
    # 2) 新格式: "in level shop put the bottle_blue into the basket"
    m = re.search(r'put the ([\w_]+) into the ([\w_]+)', text)
    if m:
        obj_key  = normalize_key(m.group(1))
        goal_key = normalize_key(m.group(2))
        obj_cn   = OBJ_CN.get(obj_key, obj_key)
        goal_cn  = OBJ_CN.get(goal_key, goal_key)
        level_cn = OBJ_CN.get(level_name.lower(), level_name)
        return f"在场景{level_cn}中将{obj_cn}放入{goal_cn}中"
    
     # 3) 打开门场景: "in level hosekeeping sence, open the fridge_left_up door."
    m = re.search(r'open the ([\w_]+) door', text)
    if m:
        # obj_key  = normalize_key(m.group(1))
        goal_key = normalize_key(m.group(1))
        # obj_cn   = OBJ_CN.get(obj_key, obj_key)
        goal_cn  = OBJ_CN.get(goal_key, goal_key)
        level_cn = OBJ_CN.get(level_name.lower(), level_name)
        return f"在{level_cn}场景中打开{goal_cn}门"
    
    # 4) 关闭门场景: "in level hosekeeping sence, close the  fridge_left_up door."
    m = re.search(r'close the ([\w_]+) door', text)
    if m:
        # obj_key  = normalize_key(m.group(1))
        goal_key = normalize_key(m.group(1))
        # obj_cn   = OBJ_CN.get(obj_key, obj_key)
        goal_cn  = OBJ_CN.get(goal_key, goal_key)
        level_cn = OBJ_CN.get(level_name.lower(), level_name)
        return f"在{level_cn}场景中关闭{goal_cn}门"
    
    # 5): "in the shop scene, pick up the niuhuangwan and scan it with the pipalu"
    m = re.search(r'pick up the ([\w_]+) and scan it with the ([\w_]+)', text)
    if m:
        obj_key  = normalize_key(m.group(1))
        goal_key = normalize_key(m.group(2))
        obj_cn   = OBJ_CN.get(obj_key, obj_key)
        goal_cn  = OBJ_CN.get(goal_key, goal_key)
        level_cn = OBJ_CN.get(level_name.lower(), level_name)
        return f"在场景{level_cn}中，拿起{obj_cn}用{goal_cn}扫描"

    
    # 6) 再试 put … into … 结构
    m = re.search(r'put\s+([\w_]+)\s+into\s+([\w_]+)', text)
    if m:
        obj_key  = normalize_key(m.group(1))
        goal_key = normalize_key(m.group(2))
        obj_cn   = OBJ_CN.get(obj_key, obj_key)
        goal_cn  = OBJ_CN.get(goal_key, goal_key)
        return f"将{obj_cn}放入{goal_cn}中"
    
    # 7) 再试 move … to … 结构
    m = re.search(r'move\s+([\w_]+)\s+to\s+([\w_]+)', text)
    if m:
        obj_key  = normalize_key(m.group(1))
        goal_key = normalize_key(m.group(2))
        obj_cn   = OBJ_CN.get(obj_key, obj_key)
        goal_cn  = OBJ_CN.get(goal_key, goal_key)
        return f"将{obj_cn}移动到{goal_cn}前"
    
    # 8) 回退：保留原文
    return instruction_en

def register_env(orcagym_addr : str, 
                 env_name : str, 
                 env_index : int, 
                 agent_names : str,
                 pico_ports : str,
                 run_mode : str, 
                 action_type : str,
                 ctrl_device : str,
                 max_episode_steps : int,
                 sample_range : float,
                 action_step : int,
                 camera_config : Dict[str, Any],
                 task_config_dict: Dict[str, Any] = None) -> str:
    orcagym_addr_str = orcagym_addr.replace(":", "-")
    env_id = env_name + "-OrcaGym-" + orcagym_addr_str + f"-{env_index:03d}"
    agent_names_list = agent_names.split(" ")
    print("Agent names: ", agent_names_list)
    pico_ports = pico_ports.split(" ")
    print("Pico ports: ", pico_ports)
    kwargs = {'frame_skip': FRAME_SKIP,   
                'reward_type': RewardType.SPARSE,
                'orcagym_addr': orcagym_addr, 
                'agent_names': agent_names_list,
                'pico_ports': pico_ports,
                'time_step': TIME_STEP,
                'run_mode': run_mode,
                'action_type': action_type,
                'ctrl_device': ctrl_device,
                'control_freq': CONTROL_FREQ,
                'sample_range': sample_range,
                'action_step': action_step,
                'camera_config': camera_config,
                'task_config_dict': task_config_dict}
    gym.register(
        id=env_id,
        entry_point=ENV_ENTRY_POINT[env_name],
        kwargs=kwargs,
        max_episode_steps= max_episode_steps,
        reward_threshold=0.0,
    )
    return env_id, kwargs


# def teleoperation_episode(env : DualArmEnv, cameras : list[CameraWrapper], rgb_size : tuple = (256, 256), action_step : int = 1):
def teleoperation_episode(env : DualArmEnv, cameras : list[CameraWrapper], dataset_writer : DatasetWriter,rgb_size : tuple = (256, 256), action_step : int = 1):
    """_summary_

    Args:
        env (DualArmEnv): envairoment instance for franka robot manipulation task
        cameras (list[CameraWrapper]): list of camera instances, (primary, secondary, wrist)
        rgb_size (tuple, optional): image size for rgb camera. Defaults to (256, 256).
        action_step (int, optional): 
            How many physics steps to take in each time the model pridiects an action. 
            
            - For the robomimic model, the model pridiects an action for every physics step.
            Witch means the model can run in realtime speed.
            
            - For the VLA model, the model pridiects an action for every 5 physics steps. 
            Means the model can only run in 1/5 realtime speed. (about 5 Hz on Nvidia 3090 GPU, as the paper reported.)

    Returns:
        _type_: _description_
    """
    env._task.spawn_scene(env) # 初始化场景
    obs, info = env.reset(seed=42)
    obs_list = {obs_key: list([]) for obs_key, obs_data in obs.items()}
    reward_list = []
    done_list = []
    info_list = []    
    terminated_times = 0
    is_success_list = []
    # camera_frames = {camera.name: [] for camera in cameras}
    camera_frame_index = []
    camera_time_stamp = {}
    timestep_list = []
    action_step_taken = 0
    saving_mp4 = False
    while True:
        start_time = datetime.now()

        action = env.action_space.sample()
        obs, reward, terminated, truncated, info = env.step(action)
        # print(f"info: {info}")

        env.render()
        task_status = info['task_status']

        action_step_taken += 1
        is_success = False
        if action_step_taken >= action_step:        
            action_step_taken = 0
            if task_status in [TaskStatus.SUCCESS, TaskStatus.FAILURE, TaskStatus.BEGIN]:
                if task_status == TaskStatus.BEGIN:
                    if saving_mp4 == False:
                        dataset_writer.set_UUIDPATH()
                        mp4_save_path = dataset_writer.get_mp4_save_path()

                        env.begin_save_video(mp4_save_path, 0)

                        saving_mp4 = True
                        camera_frame_index.append(env.get_current_frame())
                        
                    else:
                        camera_frame_index.append(env.get_current_frame())

                if task_status == TaskStatus.SUCCESS:
                    current_frame = env.get_current_frame()
                    camera_frame_index.append(current_frame)
                    print(f"camera_frame_index: {camera_frame_index[-1]}")
                    is_success = env._task.is_success(env)
                    is_success_list.append(is_success)
                    # 只有在相机帧索引有效时才尝试获取时间戳
                    if is_success and current_frame >= 0:
                        try:
                            time_stamp_dict = env.get_camera_time_stamp(camera_frame_index[-1])
                            for camera_name, time_list in time_stamp_dict.items():
                                try:
                                    camera_time_stamp[camera_name] = [time_list[index] for index in camera_frame_index]
                                except IndexError:
                                    print(f"Warning: IndexError when processing camera {camera_name} timestamps")
                                    # 不将is_success设置为False，因为任务本身是成功的
                        except Exception as e:
                            print(f"Warning: Failed to get camera timestamps: {e}")
                            # 不将is_success设置为False，因为任务本身是成功的
                    elif is_success and current_frame < 0:
                        print("Warning: Camera frame index is invalid, but task is successful. Proceeding without camera timestamps.")
                    
                    env.stop_save_video()
                    saving_mp4 = False
                    # 任务成功时不删除路径，即使is_success为False（可能是相机问题导致的）
                    # 只有在任务状态为FAILURE时才删除路径
                    if task_status == TaskStatus.FAILURE:
                        dataset_writer.remove_path()
                elif task_status == TaskStatus.FAILURE:

#                     env.stop_save_video()
#                     saving_mp4 = False

                    # env.unwrapped.stop_save_video()
                    env.stop_save_video()
                    #通过os接口删除路径

                    dataset_writer.remove_path()
                  #      print(f"目录及其内容已删除: {rmpath}")
           
                    saving_mp4 = False              

                for obs_key, obs_data in obs.items():
                    obs_list[obs_key].append(obs_data)
                    
                reward_list.append(reward)
                done_list.append(0 if not terminated else 1)
                info_list.append(info)
                terminated_times = terminated_times + 1 if terminated else 0
                timestep_list.append(info['time_step'])
                

        if terminated_times >= 5 or truncated or is_success:
            return obs_list, reward_list, done_list, info_list, camera_frame_index, timestep_list,is_success_list, camera_time_stamp

        elapsed_time = datetime.now() - start_time
        if elapsed_time.total_seconds() < REALTIME_STEP:
            time.sleep(REALTIME_STEP - elapsed_time.total_seconds())
                            
def user_comfirm_save_record(task_result, currnet_round, teleoperation_rounds):
    while True:
        user_input = input(f"Round {currnet_round} / {teleoperation_rounds}, Task is {task_result}! Do you want to save the record? y(save), n(ignore), e(ignore & exit): ")
        if user_input == 'y':
            return True, False
        elif user_input == 'n':
            return False, False
        elif user_input == 'e':
            return False, True
        else:
            print("Invalid input! Please input 'y', 'n' or 'e'.")

def append_task_info_json(
        json_path,
        episode_id: str,
        level_name: str,
        sub_scene_name: str,
        language_instruction_cn: str,
        language_instruction_en: str,
        action_config: list[dict],
        data_gen_mode: str = "simulation",
        sn_code: str = "A2D0001AB00029",
        sn_name: str = "青龙"):
    """把一条 episode 的元信息追加到同一个 task_info.json 里。"""
    lock = FileLock(json_path + ".lock", timeout=20)
    try:
        with lock:
            # 确保目录存在
            dirpath = os.path.dirname(json_path)
            if dirpath:
                os.makedirs(dirpath, exist_ok=True)

            # 读旧数据（文件不存在或损坏都当作空列表）
            try:
                with open(json_path, 'r', encoding='utf-8') as f:
                    all_eps = json.load(f)
            except (FileNotFoundError, json.JSONDecodeError):
                all_eps = []

            # 构造新 episode
            scene_key = "shop" if level_name == "jiazi" else level_name
            init_cn, init_en = INIT_SCENE_TEXT.get(scene_key, ("", ""))
            episode = {
                "episode_id": episode_id,
                "scene_name": scene_key.title(),
                "sub_scene_name": sub_scene_name,
                "init_scene_text": init_cn,
                "english_init_scene_text": init_en,
                "task_name": language_instruction_cn,
                "english_task_name": language_instruction_en,
                "data_type": "常规",
                "episode_status": "approved",
                "data_gen_mode": data_gen_mode,
                "sn_code": sn_code,
                "sn_name": sn_name,
                "label_info": {
                    "action_config": action_config,
                    "key_frame": []
                }
            }
            all_eps.append(episode)

            # 原子写入
            tmp_path = json_path + ".tmp"
            with open(tmp_path, 'w', encoding='utf-8') as f:
                json.dump(all_eps, f, ensure_ascii=False, indent=4)
            os.replace(tmp_path, json_path)

    except Timeout:
        raise RuntimeError(f"无法获得文件锁：{json_path}.lock")

def add_demo_to_dataset(dataset_writer : DatasetWriter,
                        obs_list, 
                        reward_list, 
                        done_list, 
                        info_list, 
                        camera_frames,
                        camera_time_stamp,
                        timestep_list, 
                        language_instruction,
                        level_name):
        
    # 只处理第一个info对象（初始状态）
    first_info = info_list[0]

    # 将objects和goals转换为简单的numpy数组格式
    def convert_to_simple_arrays(structured_array):
        """将结构化数组转换为简单的numpy数组，便于HDF5存储"""
        if len(structured_array) == 0:
            return np.array([])
        
        # 提取所有数值数据并展平
        data_list = []
        for field_name in structured_array.dtype.names:
            if structured_array.dtype[field_name].kind == 'U':  # Unicode string
                # 跳过字符串字段，只保留数值数据
                continue
            else:
                data = structured_array[field_name]
                if data.dtype.kind in ['f', 'i', 'u']:  # 数值类型
                    data_list.append(data.flatten())
        
        if data_list:
            return np.concatenate(data_list).astype(np.float32)
        else:
            return np.array([])
    
    objects_array = convert_to_simple_arrays(first_info['object'])
    goals_array = convert_to_simple_arrays(first_info['goal'])

    dataset_writer.add_demo_data({
        'states': np.array([np.concatenate([info["state"]["qpos"], info["state"]["qvel"]]) for info in info_list], dtype=np.float32),
        'actions': np.array([info["action"] for info in info_list], dtype=np.float32),
        'objects': objects_array,
        'goals': goals_array,
        'rewards': np.array(reward_list, dtype=np.float32),
        'dones': np.array(done_list, dtype=np.int32),
        'obs': obs_list,
        'camera_frames': camera_frames,
        'camera_time_stamp': camera_time_stamp,
        'timesteps': np.array(timestep_list, dtype=np.float32),
        'timestamps': np.array([info["time_stamp"] for info in info_list], dtype=np.uint64),
        'language_instruction': language_instruction
    })
    # 1) 拿到 level_name
    lvl = level_name.decode() if isinstance(level_name, (bytes, bytearray)) else level_name
    
    # 2) 映射出 scene_name / sub_scene_name
    scene_name, sub_scene_name = SCENE_SUBSCENE_MAPPING.get(
        lvl, (lvl.title(), lvl.title()+"_Operation")
    )
    
    # 3) 构造 JSON 目录 (放在 scene 目录下)
    #    假设 dataset_writer.basedir = "records_tmp/yaodian/dual_arm_… .hdf5"
    scene_dir = os.path.join(
        dataset_writer.basedir #,  # parent of HDF5 filename
         #lvl  # 目录名和你的一级 level 保持一致，例如 "shop", "yaodian"…
    )
    os.makedirs(scene_dir, exist_ok=True)
    
    # 4) JSON 路径
    json_fname = f"{scene_name}-{sub_scene_name}.json"
    json_path = os.path.join(scene_dir, json_fname)
   
    
    # 5) 时间戳、帧信息照旧
    start_frame = 0
    max_len     = max(len(frames) for frames in camera_frames.values()) if isinstance(camera_frames, dict) else len(camera_frames)
    end_frame   = camera_frames[-1]
    from datetime import datetime, timezone, timedelta
    CST = timezone(timedelta(hours=8))
    timestamp_cst = datetime.now(CST).isoformat()
    
    # 6) 生成中／英文指令
    lang_en = language_instruction
    if isinstance(lang_en, (bytes, bytearray)):
        lang_en = lang_en.decode('utf-8', errors='ignore')
    lang_cn = eng2cn(lang_en, level_name)
    
    action_config = [{
        "start_frame": start_frame,
        "end_frame":   end_frame,
        "timestamp_utc": timestamp_cst,
        "skill": "pick_and_place",
        "action_text":       lang_cn,
        "english_action_text": lang_en
    }]
    
    # 7) 调用追加
    append_task_info_json(
        json_path=json_path,
        episode_id=dataset_writer.experiment_id,
        level_name=lvl,
        sub_scene_name=sub_scene_name,
        language_instruction_cn=lang_cn,
        language_instruction_en=lang_en,
        action_config=action_config,
        data_gen_mode="simulation",
    )
    uuid_dir   = dataset_writer.get_UUIDPath()
    dump_static_camera_params(uuid_dir, lvl)

def do_teleoperation(env, 
                     dataset_writer : DatasetWriter, 
                     teleoperation_rounds : int, 
                     cameras : list[CameraWrapper], 
                     rgb_size : tuple = (256, 256),
                     action_step : int = 1,
                     output_video : bool = True):    
    
    current_round = 1
    
    for camera in cameras:
        camera.start()

    while True:
        obs_list, reward_list, done_list, info_list, camera_frame_index, timestep_list, is_success_list, camera_time_stamp = teleoperation_episode(
        env, cameras, dataset_writer, rgb_size, action_step
    )
    
        last_done = (len(done_list) > 0 and done_list[-1] == 1)
        # 只要任务完成就保存数据，不依赖于is_success检查
        save_record = last_done
        task_result = "Success" if save_record else "Failed"
        exit_program = False
        
      #  print(f"info_list: {info_list}")
    
        if save_record:
            print(f"Round {current_round} / {teleoperation_rounds}, Task is {task_result}!")
            current_round += 1

            add_demo_to_dataset(dataset_writer, obs_list, reward_list, done_list, info_list, camera_frame_index, camera_time_stamp, timestep_list, info_list[0]["language_instruction"],level_name = env._task.level_name)
            uuid_path = dataset_writer.get_UUIDPath()
            camera_name_list = []
            for camera_name in camera_time_stamp.keys():
                if camera_name.endswith("_color"):
                    camera_name_list.append(camera_name.replace("_color", ""))
            # unitCheack = BasicUnitChecker(uuid_path, camera_name_list, "proprio_stats.hdf5")
            # ret, _ = unitCheack.check()
            # if ret != ErrorType.Qualified:
            #     dataset_writer.remove_path()
        if exit_program or current_round > teleoperation_rounds:
            break
        
    for camera in cameras:
        camera.stop()

def playback_episode(env : DualArmEnv, 
                     action_list : list[np.ndarray], 
                     done_list : list[int],
                     action_step : int = 1,
                     realtime : bool = False):
    for i in range(len(action_list)):

        action = action_list[i]
        last_action = action_list[i - 1] if i > 0 else action
        # 插值生成一组动作序列，长度为 action_step
        action_chunk = np.array([last_action + (action - last_action) * (i / action_step) for i in range(action_step)])
        # print("Playback Action: ", action_chunk)

        done = done_list[i]
        
        for i in range(action_step):
            if realtime:
                start_time = datetime.now()

            obs, reward, terminated, truncated, info = env.step(action_chunk[i])
            env.render()

            if realtime:
                elapsed_time = datetime.now() - start_time
                if elapsed_time.total_seconds() < REALTIME_STEP:
                    time.sleep(REALTIME_STEP - elapsed_time.total_seconds())

        if done:
            print("Episode done!")
            return

        if terminated or truncated:
            print("Episode terminated!")
            return
    
    print("Episode tunkated!")

def all_object_joints_exist(cfg, model) -> bool:
    """
    检查 cfg 中定义的 object_joints 是否都存在于 model 的 joint 字典中。
    """
    expected_joints = cfg.get("object_joints", [])
    model_joints = model.get_joint_dict().keys()

    missing = [j for j in expected_joints if not any(j in mj for mj in model_joints)]
    
    if missing:
        print("[Check] Missing joints:")
        for j in missing:
            print(f"  - {j} (expected but not found in model)")
        return False

    return True

def spawn_scene(env: DualArmEnv, task_config: Dict[str, Any]) -> None:
    global _light_counter

    env._task.load_config(task_config)

    # 初次初始化，什么都没有，先把object创建出来
    if not all_object_joints_exist(task_config, env.model):
        env._task.publish_scene()   # 清空当前缓存
        env._task.generate_actors(random_actor=False)
        env._task.publish_scene()
        time.sleep(1)  # 等待场景加载完成


    # 满足light切换条件，从新创建一次场景
    if env._config.get("random_light", False) and (_light_counter % _LIGHT_SWITCH_PERIOD == 0):
        actor_idxs = env._task.generate_actors(random_actor=False)
        light_idxs = env._task.generate_lights()
        env._task.publish_scene()

        # 先publish scene，确保所有actor和light都被创建，然后再设置
        if env._task.random_actor_color:
            for idx in actor_idxs:
                env._task.set_actor_material(env._task.actors[idx])

        for idx in light_idxs:
            env._task.set_light_info(env._task.lights[idx])
        env.mj_forward()
        env.render()
        time.sleep(1)  # 等待场景加载完成

    _light_counter += 1

# todo： objects 和 goal的数据结构有改变，需要重新处理
def reset_playback_env(env: DualArmEnv, demo_data, sample_range=0.0):
    if "objects" in demo_data:
        env.objects = demo_data["objects"]        # 结构化物体信息，reset_model会用

    # 场景初始化
    env._task.spawn_scene(env)
    obs, info = env.reset(seed=42)

    # 如果需要重新采样目标物体位置
    if sample_range > 0.0:
        replace_objects(env, demo_data) # 先还原，获得物品的原始旋转值
        resample_objects(env, demo_data, sample_range)
    else:
        replace_objects(env, demo_data)

    obs = env._get_obs().copy()
    info = {
        "object": demo_data['objects'],
        "goal":   demo_data.get('goals')
    }
    return obs, info


def get_target_object_and_goal(demo: dict) -> tuple:
    lang_instr = demo.get("language_instruction", b"")
    if isinstance(lang_instr, (bytes, bytearray)):
        lang_instr = lang_instr.decode("utf-8")
    obj_match = re.search(r'object:\s*([^\s]+)', lang_instr)
    goal_match = re.search(r'goal:\s*([^\s]+)', lang_instr)
    target_obj = obj_match.group(1) if obj_match else None
    target_goal = goal_match.group(1) if goal_match else None

    return target_obj, target_goal

def replace_objects(env: DualArmEnv, demo_data: dict) -> None:
    env.replace_objects(demo_data['objects'])
    if 'goals' in demo_data and demo_data['goals'] is not None:
        env.replace_goals(demo_data['goals'])
    env.mj_forward()

def resample_objects(env: DualArmEnv, demo: dict, sample_range: float) -> None:
    env._task.load_config(env._config)
    target_obj, _ = get_target_object_and_goal(demo)
    if target_obj:
        target_obj_joint_name = ""
        target_obj_position = np.array([0.0, 0.0, 0.0], dtype=np.float32)
        demo_objects = demo.get("objects", [])
        for obj in demo_objects:
            joint_name = obj["joint_name"].decode('utf-8') if isinstance(obj["joint_name"], (bytes, bytearray)) else obj["joint_name"]
            if target_obj in joint_name:
                target_obj_joint_name = joint_name
                target_obj_position = np.array(obj["position"], dtype=np.float32)
                break

        resample_success = False
        target_obj_position_delta = 0
        for _ in range(100):  # 尝试100次
            env._task.random_objs_and_goals(env, random_rotation=False, target_obj_joint_name=target_obj_joint_name)
            target_obj_joint_qpos = env.query_joint_qpos([target_obj_joint_name])[target_obj_joint_name]
            target_obj_position_delta = np.linalg.norm(target_obj_joint_qpos[:2] - target_obj_position[:2])
            # print(f"[Info] Resampling target object {target_obj} position: {target_obj_joint_qpos[:2]} (delta: {target_obj_position_delta})")
            if target_obj_position_delta < sample_range:
                resample_success = True
                break

        if not resample_success:
            print(f"Warning: Failed to resample target object {target_obj} within range {sample_range}. Using original position.")

        print("[Info] Resampling target object", target_obj, "to delta:", target_obj_position_delta)

def calculate_transform_matrix(a, b, a1, b1):
    """
    计算从向量[a, b]到[a1, b1]的变换矩阵
    返回一个2x2 NumPy数组
    
    注意: 当输入向量为零向量时处理特殊情况
    """
    # 计算分母 D = a² + b²
    D = a**2 + b**2
    
    if D == 0:  # 输入向量是零向量
        if a1 == 0 and b1 == 0:
            # 返回单位矩阵 (但这不是唯一解)
            return np.eye(2)
        else:
            # 线性变换必须保持零向量不变
            # raise ValueError("输入向量是零向量但输出向量非零 - 无效线性变换")
            return np.eye(2)
    
    # 使用公式计算矩阵元素
    m11 = (a*a1 + b*b1) / D
    m12 = (b*a1 - a*b1) / D  # 注意这等同于公式中的 -(a*b1 - b*a1)/D
    
    # 简化形式创建矩阵 [m11, m12; -m12, m11]
    return np.array([[m11, m12], [-m12, m11]])

def apply_transform(matrix, x, y):
    """
    应用变换矩阵到向量[x, y]
    返回变换后的向量[x1, y1]
    """
    # 将输入向量转换为列向量
    vector = np.array([[x], [y]])
    
    # 应用矩阵乘法
    transformed = matrix @ vector
    
    # 返回结果为一维数组
    return transformed.flatten()

def _get_noscale_action_list(env: DualArmEnv, action_list: np.ndarray) -> np.ndarray:
    noscale_action_list = []
    for action in action_list:
        noscale_action = env.denormalize_action(action, env.env_action_range_min, env.env_action_range_max)
        noscale_action_list.append(noscale_action)
    return np.array(noscale_action_list)

def _get_scaled_action_list(env: DualArmEnv, noscale_action_list: np.ndarray) -> np.ndarray:
    scaled_action_list = []
    for noscale_action in noscale_action_list:
        scaled_action = env.normalize_action(noscale_action, env.env_action_range_min, env.env_action_range_max)
        scaled_action_list.append(scaled_action)
    return np.array(scaled_action_list)

def _get_eef_xy_from_action(env: DualArmEnv, noscale_action_list: np.ndarray) -> np.ndarray:
    xy_list = np.zeros((0, 4), dtype=np.float32)
    robot: DualArmRobot = env._agents[env._agent_names[0]]
    for noscale_action in noscale_action_list:
        left_eef_action = robot._action_B_to_action(noscale_action[:6])
        right_eef_action = robot._action_B_to_action(noscale_action[14:20])
        xy = np.concatenate([left_eef_action[:2], right_eef_action[:2]]).flatten()
        xy_list = np.concatenate([xy_list, [xy]])

    return xy_list

def _set_eef_xy_to_action(env: DualArmEnv, noscale_action_list: np.ndarray, xy_list: np.ndarray) -> np.ndarray:
    robot: DualArmRobot = env._agents[env._agent_names[0]]
    
    for i, noscale_action in enumerate(noscale_action_list):
        left_eef_action = robot._action_B_to_action(noscale_action[:6])
        right_eef_action = robot._action_B_to_action(noscale_action[14:20])
        left_eef_action[:2] = xy_list[i][:2]
        right_eef_action[:2] = xy_list[i][2:]
        noscale_action[:6] = robot._action_to_action_B(left_eef_action)
        noscale_action[14:20] = robot._action_to_action_B(right_eef_action)

    return noscale_action_list

def _get_target_xy(env: DualArmEnv, demo_data: dict, target_obj: str) -> tuple[np.ndarray, np.ndarray]:
    target_obj_full_name = env.body(target_obj) + "_joint"
    env_objects = json.loads(env._task.get_objects_info(env))
    demo_objects = demo_data["objects"]

    # 版本兼容：新版本的 objects 是一个 JSON 字符串或者从hdf5文件中读取的json， 老版本是ndarray
    def _get_object_info(obj_joint_name: str, objects) -> dict:
        if (demo_objects.shape == () and demo_objects.dtype == "object"):
            json_str = demo_objects[()]
            json_data = json.loads(json_str)
            for object, object_info in json_data.items():
                if env.joint(object_info['joint_name']) == obj_joint_name:
                    return {
                        "joint_name": env.joint(object_info['joint_name']),
                        "position": np.array(object_info['position'], dtype=np.float32),
                        "orientation": np.array(object_info['orientation'], dtype=np.float32)
                    }

        else:
            arr = demo_objects
            for entry in arr:
                name = str(entry['joint_name'], encoding="utf-8")
                if name == obj_joint_name:
                    return {
                        "joint_name": name,
                        "position": np.array(entry['position'], dtype=np.float32),
                        "orientation": np.array(entry['orientation'], dtype=np.float32)
                    }

        return {}
    
    target_obj_demo_info = _get_object_info(target_obj_full_name, demo_objects)

    demo_xy = target_obj_demo_info.get("position", np.zeros(3, dtype=np.float32))[:2]
    env_xy = env_objects[target_obj]["position"][:2]
    return demo_xy, env_xy

def _find_closest_step(xy_list: np.ndarray, target_xy: np.ndarray) -> int:
    closest_step_id = 0
    closest_step_delta = float('inf')
    for i, xy in enumerate(xy_list):
        left_eef_xy = xy[:2]
        right_eef_xy = xy[2:]
        delta_left_xy = np.linalg.norm(left_eef_xy - target_xy)
        delta_right_xy = np.linalg.norm(right_eef_xy - target_xy)
        if delta_left_xy < 0.05 or delta_right_xy < 0.05:
            # print("Env: close to target object, left: ", delta_left_xy, " right: ", delta_right_xy, "in action: ", i)
            if delta_left_xy < closest_step_delta or delta_right_xy < closest_step_delta:
                closest_step_delta = min(delta_left_xy, delta_right_xy)
                closest_step_id = i

    return closest_step_id

def _transform_xy_list(source_xy_list: np.ndarray, source_xy: np.ndarray, target_xy: np.ndarray, skip_steps: int) -> np.ndarray:
    transform_matrix = calculate_transform_matrix(
        source_xy[0], source_xy[1], target_xy[0], target_xy[1]
    )
    print("Transform Matrix: ", transform_matrix)

    target_xy_list = np.zeros_like(source_xy_list)
    for i, xy in enumerate(source_xy_list):
        if i <= skip_steps:
            target_xy_list[i] = xy
            continue
        left_eef_xy = xy[:2]
        right_eef_xy = xy[2:]
        transformed_left_xy = apply_transform(transform_matrix, left_eef_xy[0], left_eef_xy[1])
        transformed_right_xy = apply_transform(transform_matrix, right_eef_xy[0], right_eef_xy[1])
        target_xy_list[i] = np.concatenate([transformed_left_xy, transformed_right_xy])

    return target_xy_list

def resample_actions(
        env: DualArmEnv, 
        demo_data: dict,
    ) -> np.ndarray:
    """
    两个关注点：1. 夹爪与物体最接近的点，是第一个目标点， 2. 样本中最后一个点是第二个目标点

    首先根据第一个目标点重采样后，与原始数据之间的变化，建立一个变换矩阵（平移+缩放）
    对所有轨迹应用变换矩阵，则夹爪可以移动到新的目标位置
    这时最终目标点也会有变化。因此计算第二个目标点的变换矩阵，
    然后对分界点后的轨迹，再次应用变换矩阵，获得第二段轨迹

    最终效果：夹爪到达新物体位置，然后再到达原轨迹中的最终位置
    """
    demo_action_list = demo_data['actions']
    target_obj = env._task.target_object

    if target_obj is None:
        print("Warning: No target object found in the demo data.")
        return demo_action_list
    
    noscale_action_list = _get_noscale_action_list(env, demo_action_list)
    demo_xy_list = _get_eef_xy_from_action(env, noscale_action_list)
    # print("Get xy_list from demo actions: ", demo_xy_list, "shape: ", demo_xy_list.shape)
    # sys.exit(0)

    demo_xy, env_xy = _get_target_xy(env, demo_data, target_obj)
    final_demo_xy = demo_xy_list[-1]
    print("Demo XY: ", demo_xy, "Env XY: ", env_xy, "Final Demo XY: ", final_demo_xy)

    closest_demo_step_id = _find_closest_step(demo_xy_list, demo_xy)
    print("Closest Demo Step ID: ", closest_demo_step_id)

    env_xy_list = _transform_xy_list(demo_xy_list, demo_xy, env_xy, 0)
    final_env_xy = env_xy_list[-1]
    print("Final Env XY: ", final_env_xy)

    closest_env_step_id = _find_closest_step(env_xy_list, env_xy)
    print("Closest Env Step ID: ", closest_env_step_id)

    env_xy_list = _transform_xy_list(env_xy_list, final_env_xy, final_demo_xy, closest_demo_step_id)
    new_final_env_xy = env_xy_list[-1]
    print("New Final Env XY: ", new_final_env_xy, "Demo XY: ", final_demo_xy, "distance: ", np.linalg.norm(new_final_env_xy - final_demo_xy))
    

    noscale_action_list = _set_eef_xy_to_action(env, noscale_action_list, env_xy_list)

    scaled_action_list = _get_scaled_action_list(env, noscale_action_list)

    return scaled_action_list
    
def do_playback(env : DualArmEnv, 
                dataset_reader : DatasetReader, 
                playback_mode : str,
                action_step : int = 1,
                realtime : bool = False):
    demo_names = dataset_reader.get_demo_names()
    if playback_mode == "loop":
        demo_name_index_list = list(range(len(demo_names)))
    elif playback_mode == "random":
        demo_name_index_list = np.random.permutation(list(range(len(demo_names))))
    else:
        print("Invalid playback mode! Please input 'loop' or 'random'.")
        return
    
    for i in demo_name_index_list:
        demo_data = dataset_reader.get_demo_data(demo_names[i])
        action_list = demo_data['actions']
        done_list = demo_data['dones']
        print("Playing back episode: ", demo_names[i], " with ", len(action_list), " steps.")
        # for i, action in enumerate(action_list):
        #     print(f"Playback Action ({i}): ", action)
        env.objects = demo_data['objects']
        reset_playback_env(env, demo_data)
        playback_episode(env, action_list, done_list, action_step, realtime)
        time.sleep(1)

def augment_episode(env : DualArmEnv, 
                    cameras : list[CameraWrapper], 
                    rgb_size : tuple,
                    demo_data : dict, 
                    noise_scale : float, 
                    sample_range : float, 
                    realtime : bool = False,
                    action_step : int = 1,
                    sync_codec: bool = False,
                    output_video: bool = True,
                    output_video_path: str = "") -> tuple:
    env._task.data = demo_data
    env._task.spawn_scene(env)
    env._task.sample_range = sample_range
    obs, _ = env.reset(seed=42)
    obs_list    = {obs_key: [] for obs_key, obs_data in obs.items()}
    reward_list = []
    done_list = []
    info_list = []    
    terminated_times = 0    
  # camera_frames = {camera.name: [] for camera in cameras}
    camera_frame_index = []
    timestep_list = []

    is_success=False

    # 轨迹增广需要重新采样动作
    if sample_range > 0.0:
        action_list = resample_actions(env, demo_data)
    else:
        action_list = demo_data['actions']

    action_index_list = list(range(len(action_list)))
    holdon_action_index_list = action_index_list[-1] * np.ones(20, dtype=int)
    action_index_list = np.concatenate([action_index_list, holdon_action_index_list]).flatten()
    original_dones = np.array(demo_data['dones'], dtype=int)
    T = len(original_dones)

    if output_video:
        print(f"video save path: {output_video_path}")
        env.begin_save_video(output_video_path, int(sync_codec))
    for i in action_index_list:
        action = action_list[i]
        last_action = action_list[i - 1] if i > 0 else action
    
        action_chunk = np.array([
            last_action + (action - last_action) * (j / action_step)
            for j in range(action_step)
        ], dtype=np.float32)
    
        if noise_scale > 0.0:
            noise = np.random.normal(0, noise_scale, action_chunk.shape)
            action_chunk += noise * np.abs(action_chunk)
    
        action_chunk = np.clip(action_chunk, -1.0, 1.0)
    
        for j in range(action_step):
            if realtime:
                start_time = datetime.now()
    
            obs, reward, terminated, truncated, info = env.step(action_chunk[j])
            terminated_times = terminated_times + 1 if terminated else 0
            timestep_list.append(env.gym.data.time)
    
            if realtime:
                env.render()
                elapsed_time = datetime.now() - start_time
                if elapsed_time.total_seconds() < REALTIME_STEP:
                    time.sleep(REALTIME_STEP - elapsed_time.total_seconds())
            else:
                env.render()

        is_success = False
        if original_dones[i] :
            is_success = env._task.is_success(env)

        global g_skip_frame
        if sync_codec and g_skip_frame < 1:
            camera_frame_index.append(env.get_next_frame())
            g_skip_frame += 1
            # print("Get next frame, sync_codec:", sync_codec, "g_skip_frame:", g_skip_frame)
        else:
            camera_frame_index.append(env.get_current_frame())
            g_skip_frame = 0
            # print("Get current frame, sync_codec:", sync_codec, "g_skip_frame:", g_skip_frame)

        for obs_key, obs_data in obs.items():
            obs_list[obs_key].append(obs_data)
            
        reward_list.append(reward)
        done_list.append(int(original_dones[i]) if i < T else 1)
        info_list.append(info)

        if terminated_times >= 5 or truncated:
            if output_video:
                env.stop_save_video()
            return obs_list, reward_list, done_list, info_list, camera_frame_index, timestep_list, is_success_list, {}

    if output_video:
        env.stop_save_video()
        # 只有在相机帧索引有效时才尝试获取时间戳
        if len(camera_frame_index) > 0 and camera_frame_index[-1] >= 0:
            try:
                time_stamp_dict = env.get_camera_time_stamp(camera_frame_index[-1] + 1)
                return obs_list, reward_list, done_list, info_list, camera_frame_index, timestep_list, is_success_list, time_stamp_dict
            except Exception as e:
                print(f"Warning: Failed to get camera timestamps at episode end: {e}")
                return obs_list, reward_list, done_list, info_list, camera_frame_index, timestep_list, is_success_list, {}
        else:
            print("Warning: Invalid camera frame index at episode end, returning empty timestamps")
            return obs_list, reward_list, done_list, info_list, camera_frame_index, timestep_list, is_success_list, {}
    else:
        return obs_list, reward_list, done_list, info_list, camera_frame_index, timestep_list, is_success_list, {}

def do_augmentation(
        env : DualArmEnv, 
        cameras : list[CameraWrapper], 
        rgb_size : tuple,                    
        original_dataset_path : str, 
        augmented_dataset_path : str, 
        augmented_noise : float, 
        sample_range : float,
        realtime : bool,
        augmented_rounds : int,
        action_step : int = 1,
        action_type : str = "",
        output_video : bool = True,
        sync_codec : bool = False
    ):
    
    print("=================>sync codec: ", sync_codec)
   # realtime = False
    
    # Copy the original dataset to the augmented dataset
    dataset_reader = DatasetReader(file_path=original_dataset_path)

    env_kwargs=dataset_reader.get_env_kwargs()
    env_kwargs["action_type"] = action_type
    dataset_writer = DatasetWriter(base_dir=os.path.dirname(augmented_dataset_path),
                                    env_name=dataset_reader.get_env_name(),
                                    env_version=dataset_reader.get_env_version(),
                                    env_kwargs=env_kwargs)

    # 构造 JSON 路径（与 add_demo_to_dataset 函数保持一致）
    # 1) 获取 level_name
    level_name = env._task.level_name
    lvl = level_name.decode() if isinstance(level_name, (bytes, bytearray)) else level_name

    # 2) 映射出 scene_name / sub_scene_name
    scene_name, sub_scene_name = SCENE_SUBSCENE_MAPPING.get(
        lvl, (lvl.title(), lvl.title()+"_Operation")
    )

    # 3) 构造 JSON 目录
    scene_dir = os.path.join(
        dataset_writer.basedir  # parent of HDF5 filename
    )
    os.makedirs(scene_dir, exist_ok=True)

    # 4) JSON 路径
    json_fname = f"{scene_name}-{sub_scene_name}.json"
    json_path = os.path.join(scene_dir, json_fname)

    for camera in cameras:
        camera.start()
    
    for round in range(augmented_rounds):    
        need_demo_count = dataset_reader.get_demo_count()
        done_demo_count = 0
            
        demo_names = dataset_reader.get_demo_names()

        for original_demo_name in demo_names:
            done = False
            trial_count = 0
            max_trials = 2
            while not done and trial_count < max_trials:
                demo_data = dataset_reader.get_demo_data(original_demo_name)
                print("Augmenting original demo: ", original_demo_name)
                language_instruction = demo_data['language_instruction']
                level_name = env._task.level_name
                dataset_writer.set_UUIDPATH()
                obs_list, reward_list, done_list, info_list\
                    , camera_frames, timestep_list, is_success, time_stamp_dict = augment_episode(env, cameras,rgb_size,
                                                                    demo_data, noise_scale=augmented_noise, 
                                                                    sample_range=sample_range, realtime=realtime, 
                                                                    action_step=action_step, sync_codec=sync_codec,
                                                                    output_video=output_video, output_video_path=dataset_writer.get_mp4_save_path())
                if  is_success:
                    camera_time_stamp = {}
                    camera_name_list = []
                    if output_video == True:
                        for camera_name, time_list in time_stamp_dict.items():
                            camera_time_stamp[camera_name] = [time_list[index] for index in camera_frames]
                            if camera_name.endswith("_color"):
                                camera_name_list.append(camera_name.replace("_color", ""))
                    add_demo_to_dataset(dataset_writer, obs_list, reward_list, done_list, info_list,
                                        camera_frames, camera_time_stamp, timestep_list, language_instruction,level_name)
                    uuid_path = dataset_writer.get_UUIDPath()
                    # unitCheack = BasicUnitChecker(uuid_path, camera_name_list, "proprio_stats.hdf5")
                    # ret, _ = unitCheack.check()
                    # if ret != ErrorType.Qualified:
                    #     trial_count += 1
                    #     print(f"ret = {ret}")
                    #     dataset_writer.remove_path()
                    #     dataset_writer.remove_episode_from_json(json_path, dataset_writer.experiment_id)
                    # else:
                    done_demo_count += 1
                    print(f"Episode done! {done_demo_count} / {need_demo_count} for round {round + 1}")
                    done = True
                else:
                    dataset_writer.remove_path()
                    print("Episode failed! Retrying...")
                    trial_count += 1
            if not done:
                print(f"Failed to augment demo {original_demo_name} after {max_trials} tries. Skipping.")
    

    dataset_writer.shuffle_demos()
    dataset_writer.finalize()       
    
    for camera in cameras:
        camera.stop()
    
def start_monitor(port=7070, project_root : str = None):
    """
    启动 monitor.py 作为子进程。
    """
    monitor_script = f"{project_root}/orca_gym/scripts/camera_monitor.py"

    # 启动 monitor.py
    # 使用 sys.executable 确保使用相同的 Python 解释器
    process = subprocess.Popen(
        [sys.executable, monitor_script, "--port", f"{port}"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE
    )
    return process

def terminate_monitor(process):
    """
    终止子进程。
    """
    try:
        if os.name != 'nt':
            # Unix/Linux: 发送 SIGTERM 给整个进程组
            os.killpg(os.getpgid(process.pid), signal.SIGTERM)
        else:
            # Windows: 使用 terminate 方法
            process.terminate()
    except Exception as e:
        print(f"终止子进程时发生错误: {e}")



def run_example(orcagym_addr : str,
                agent_names : str,
                pico_ports : str,
                record_path : str,
                run_mode : str,
                action_type : str,
                action_step : int,
                algo_config : str,
                ctrl_device : str,
                max_episode_steps : int,
                playback_mode : str,
                rollout_times : int,
                ckpt_path : str,
                augmented_noise : float,
                augmented_rounds : int,
                teleoperation_rounds : int,
                sample_range : float,
                realtime_playback : bool,
                current_file_path : str,
                task_config : str,
                augmentation_path : str,
                output_video : bool,
                sync_codec : bool):
    try:
        print("simulation running... , orcagym_addr: ", orcagym_addr)

        if run_mode == "playback":
            dataset_reader = DatasetReader(file_path=record_path)
            print("kwargs: ", dataset_reader.get_env_kwargs())
            camera_config = dataset_reader.get_env_kwargs()["camera_config"]
            action_step = dataset_reader.get_env_kwargs()["action_step"]
            demo_action_type = dataset_reader.get_env_kwargs()["action_type"]
            env_name = dataset_reader.get_env_name()
            env_name = env_name.split("-OrcaGym-")[0]
            env_index = 0
            if task_config is None:
               task_config_dict = {}
            else:
                with open(task_config, 'r') as f:
                    task_config_dict = yaml.safe_load(f)
            env_id, kwargs = register_env(orcagym_addr, env_name, env_index, agent_names, pico_ports, RunMode.POLICY_NORMALIZED, demo_action_type, ctrl_device, max_episode_steps, sample_range, action_step, camera_config, task_config_dict)
            print("Registered environment: ", env_id)

            env = gym.make(env_id)
            env = env.unwrapped
            print("Starting simulation...")
            do_playback(env, dataset_reader, playback_mode, action_step, realtime_playback)

        elif run_mode == "teleoperation":
            env_name = ENV_NAME
            env_index = 0
            camera_config = CAMERA_CONFIG

            if task_config is None:
                task_config_dict = {}
            else:
                with open(task_config, 'r') as f:
                    task_config_dict = yaml.safe_load(f)

            env_id, kwargs = register_env(orcagym_addr, env_name, env_index, agent_names, pico_ports, RunMode.TELEOPERATION, action_type, ctrl_device, max_episode_steps, sample_range, action_step, camera_config, task_config_dict)
            print("Registered environment: ", env_id)

            env = gym.make(env_id)
            env = env.unwrapped
            print("Starting simulation...")
            kwargs["run_mode"] = RunMode.POLICY_NORMALIZED  # 此处用于训练的时候读取

            if RGB_SIZE is None:
                cameras = []
            else:
                cameras = [CameraWrapper(name=camera_name, port=camera_port) for camera_name, camera_port in camera_config.items()]


#             dataset_writer = DatasetWriter(file_path=record_path,
#                                         env_name=env_id,
#                                         env_version=env.get_env_version(),
#                                         env_kwargs=kwargs)


            # dataset_writer = DatasetWriter(file_path=record_path,
            #                             env_name=env_id,
            #                             env_version=env.unwrapped.get_env_version(),
            #                             env_kwargs=kwargs)
            dataset_writer = DatasetWriter(
                                    base_dir=os.path.dirname(record_path),           # 用 record_path 的父目录
                                    env_name=env_id,
                                    env_version=env.unwrapped.get_env_version(),
                                    env_kwargs=kwargs)



            do_teleoperation(env, dataset_writer, teleoperation_rounds,
                                                 cameras=cameras, rgb_size=RGB_SIZE, action_step=action_step,output_video = output_video)
            dataset_writer.shuffle_demos()
            dataset_writer.finalize()

        elif run_mode == "imitation":
            dataset_reader = DatasetReader(file_path=record_path)
            env_name = dataset_reader.get_env_name()
            camera_config = dataset_reader.get_env_kwargs()["camera_config"]
            action_step = dataset_reader.get_env_kwargs()["action_step"]
            demo_action_type = dataset_reader.get_env_kwargs()["action_type"]
            env_name = env_name.split("-OrcaGym-")[0]
            env_index = 0
            env_id, kwargs = register_env(orcagym_addr, env_name, env_index, agent_names, pico_ports, RunMode.POLICY_NORMALIZED, demo_action_type, ctrl_device, max_episode_steps, sample_range, action_step, camera_config)
            print("Registered environment: ", env_id)

            # env = gym.make(env_id)
            # print("Starting simulation...")

            now = datetime.now()
            formatted_now = now.strftime("%Y-%m-%d_%H-%M-%S")
            output_dir = f"{current_file_path}/trained_models_tmp/train_temp_dir_{formatted_now}"
            train_policy(config=algo_config, algo=None, dataset=record_path, name=None, output_dir=output_dir, debug=False)

        elif run_mode == "rollout":
            ckpt_dict = maybe_dict_from_checkpoint(ckpt_path=ckpt_path)

            # metadata from model dict to get info needed to create environment
            env_meta = ckpt_dict["env_metadata"]
            env_name = env_meta["env_name"]
            env_name = env_name.split("-OrcaGym-")[0]
            env_index = 0

            env_kwargs = env_meta["env_kwargs"]
            camera_config = env_kwargs["camera_config"]
            sample_range = env_kwargs["sample_range"]
            action_step = env_kwargs["action_step"]
            demo_action_type = env_kwargs["action_type"]

            env_id, kwargs = register_env(orcagym_addr, env_name, env_index, agent_names, pico_ports, RunMode.POLICY_NORMALIZED, demo_action_type, ctrl_device, max_episode_steps, sample_range, action_step, camera_config)
            print("Registered environment: ", env_id)

            env, policy = create_env(ckpt_path)

            for i in range(rollout_times):
                stats = run_rollout(
                    policy=policy,
                    env=env,
                    horizon=int(max_episode_steps / action_step),
                    render=True,
                    realtime_step=REALTIME_STEP
                )
                print(stats)
        elif run_mode == "augmentation":
            dataset_reader = DatasetReader(file_path=record_path)
            env_name = dataset_reader.get_env_name()
            camera_config = dataset_reader.get_env_kwargs()["camera_config"]
            action_step = dataset_reader.get_env_kwargs()["action_step"]
            demo_action_type = dataset_reader.get_env_kwargs()["action_type"]

            if demo_action_type in [ActionType.END_EFFECTOR_OSC, ActionType.JOINT_MOTOR] and action_type == ActionType.JOINT_POS:
                raise ValueError(f"Augmentation mode: Action type {action_type} is conflicting with demo action type {demo_action_type}.")
            elif demo_action_type in [ActionType.END_EFFECTOR_IK, ActionType.JOINT_POS] and action_type == ActionType.JOINT_MOTOR:
                raise ValueError(f"Augmentation mode: Action type {action_type} is conflicting with demo action type {demo_action_type}.")
            
            if sample_range > 0.0 and action_type not in [ActionType.END_EFFECTOR_OSC, ActionType.END_EFFECTOR_IK]:
                raise ValueError(f"Augmentation mode: Action type {action_type} does not support sample range. Please use 'end_effector_osc' or 'end_effector_ik'.")

            env_name = env_name.split("-OrcaGym-")[0]
            env_index = 0
            if task_config is None:
               task_config_dict = {}
            else:
                with open(task_config, 'r') as f:
                    task_config_dict = yaml.safe_load(f)
            env_id, kwargs = register_env(orcagym_addr, env_name, env_index, agent_names, pico_ports, RunMode.POLICY_NORMALIZED, action_type, ctrl_device, max_episode_steps, sample_range, action_step, camera_config,task_config_dict)
            print("Registered environment: ", env_id)

            env = gym.make(env_id)
            env = env.unwrapped
            print("Starting simulation...")

            now = datetime.now()
            formatted_now = now.strftime("%Y-%m-%d_%H-%M-%S")
           # agumented_dataset_file_path = f"{current_file_path}/augmented_datasets_tmp/augmented_dataset_{formatted_now}.hdf5"
            agumented_dataset_file_path = f"{augmentation_path}/augmented_dataset_{formatted_now}.hdf5"


            if RGB_SIZE is None:
                cameras = []
            else:
                cameras = [CameraWrapper(name=camera_name, port=camera_port) for camera_name, camera_port in camera_config.items()]

         #   do_augmentation(env, cameras, RGB_SIZE, record_path, agumented_dataset_file_path, augmented_noise, sample_range, augmented_rounds, action_step, output_video)
            do_augmentation(env, cameras, RGB_SIZE, record_path, agumented_dataset_file_path, augmented_noise, sample_range, realtime_playback, augmented_rounds, action_step, action_type, output_video, sync_codec)
            print("Augmentation done! The augmented dataset is saved to: ", agumented_dataset_file_path)
        else:
            print("Invalid run mode! Please input 'teleoperation' or 'playback'.")

    finally:
        print("Simulation stopped")
        # if run_mode == "teleoperation":
        if run_mode == "teleoperation" and 'dataset_writer' in locals():
            dataset_writer.finalize()
        if 'env' in locals():
            env.close()


def _get_algo_config(algo_name):
    if algo_name == "bc":
        return ["config/bc.json"]
    elif algo_name == "bc_transformer":
        return ["config/bc_transformer.json"]
    elif algo_name == "openpi":
        return ["openpi"]

    elif algo_name == "all":
        return ["config/bc.json",

                "config/bc_transformer.json"]
    else:
        raise ValueError(f"Invalid algorithm name: {algo_name}")


def run_dual_arm_sim(args, project_root : str = None, current_file_path : str = None):
    orcagym_addr = args.orcagym_address
    agent_names = args.agent_names
    pico_ports = args.pico_ports
    record_time = args.record_length
    record_path = args.dataset
    playback_mode = args.playback_mode
    run_mode = args.run_mode
    action_type = args.action_type
    action_step = args.action_step
    task_config = args.task_config
    algo = args.algo
    rollout_times = args.rollout_times
    ckpt_path = args.model_file
    augmented_noise = args.augmented_noise
    augmented_rounds = args.augmented_rounds
    teleoperation_rounds = args.teleoperation_rounds
    sample_range = args.sample_range
   # realtime_playback = args.realtime_playback
    level = args.level
    augmented_path = ''
    withvideo = True if args.withvideo == 'True' else False
    realtime_playback = True if args.realtime_playback == 'True' else False
    sync_codec = True if args.sync_codec == 'True' else False

    assert record_time > 0, "The record time should be greater than 0."
    assert teleoperation_rounds > 0, "The teleoperation rounds should be greater than 0."
    assert sample_range >= 0.0, "The sample range should be greater than or equal to 0."
    assert augmented_noise >= 0.0, "The augmented noise should be greater than or equal to 0."
    assert augmented_rounds > 0, "The augmented times should be greater than 0."

    create_tmp_dir("records_tmp")
    create_tmp_dir("trained_models_tmp")
    create_tmp_dir("augmented_datasets_tmp")

    algo_config = _get_algo_config(algo) if run_mode == "imitation" else ["none_algorithm"]

    if run_mode == "teleoperation":
        if record_path is None:
            now = datetime.now()
            formatted_now = now.strftime("%Y-%m-%d_%H-%M-%S")
            level_dir = os.path.join(current_file_path, "records_tmp", level)
            os.makedirs(level_dir, exist_ok=True)
            record_path = os.path.join(level_dir, f"dual_arm_{formatted_now}.hdf5")
            print(f"Auto-generated record path: {record_path}")
        if action_type not in [ActionType.END_EFFECTOR_OSC, ActionType.END_EFFECTOR_IK]:
            raise ValueError(f"Action type {action_type} is not supported in teleoperation mode. Please input 'end_effector_osc', 'end_effector_ik'.")
    if run_mode == "imitation" or run_mode == "playback" or run_mode == "augmentation":
        if record_path is None:
            raise ValueError("Please input the record file path.")
        else:
            augmented_path = os.path.join(current_file_path, "augmented_datasets_tmp", level)

    if run_mode == "rollout":
        if ckpt_path is None:
            raise ValueError("Please input the model file path.")
    if run_mode not in ["teleoperation", "playback", "imitation", "rollout", "augmentation"]:
        raise ValueError("Invalid run mode! Please input 'teleoperation', 'playback', 'imitation', 'rollout' or 'augmentation'.")
    if action_type not in [ActionType.END_EFFECTOR_OSC, ActionType.END_EFFECTOR_IK, ActionType.JOINT_POS, ActionType.JOINT_MOTOR]:
        raise ValueError(f"Invalid action type! Please input 'end_effector_osc', 'end_effector_ik', 'joint_pos' or 'joint_motor'.")

    if args.ctrl_device == 'vr':
        ctrl_device = ControlDevice.VR
    else:
        raise ValueError("Invalid control device! Please input 'xbox' or 'keyboard'.")

    max_episode_steps = int(record_time / REALTIME_STEP)
    print(f"Run episode in {max_episode_steps} steps as {record_time} seconds.")

    # 启动 Monitor 子进程
    # ports = [
    #     # 7070, 7080, 7090,        # Agent1
    #     # 8070, 8080, 8090,        # Agent2
    #     7070, 7090,        # Agent1
    # ]
    # monitor_processes = []
    # for port in ports:
    #     process = start_monitor(port=port, project_root=project_root)
    #     monitor_processes.append(process)

    for config in algo_config:
        run_example(orcagym_addr,
                    agent_names,
                    pico_ports,
                    record_path,
                    run_mode,
                    action_type,
                    action_step,
                    config,
                    ctrl_device,
                    max_episode_steps,
                    playback_mode,
                    rollout_times,
                    ckpt_path,
                    augmented_noise,
                    augmented_rounds,
                    teleoperation_rounds,
                    sample_range,
                    realtime_playback,
                    current_file_path,
                    task_config=task_config,
                    augmentation_path=augmented_path,
                    output_video=withvideo,
                    sync_codec=sync_codec
                    )

    # # 终止 Monitor 子进程
    # for process in monitor_processes:
    #     terminate_monitor(process)